/* --------------------------------------------------------------------------
 * WinHugs.c:	Jos Enrique Gallardo Ruiz, April 1995
 *		With mods by mpj for Hugs version 1.3, August 1996
 *
 * This file contains functions for a MS-Windows GUI for Hugs
 * ------------------------------------------------------------------------*/

/* --------------------------------------------------------------------------
 * Include standard non-GUI interpreter front end, and Windows includes
 * ------------------------------------------------------------------------*/

#include "hugs.c"

#include <ctl3d.h>
#include <commdlg.h>
#include <dir.h>
#include <stdarg.h>

/* --------------------------------------------------------------------------
 * Local function prototypes:
 * ------------------------------------------------------------------------*/

static HBITMAP local CreateStatusBar	Args((HWND));
static VOID    local DoAbout 	   	Args((Void));
static VOID    local DoCommand 	   	Args((HWND, UINT, WPARAM, LPARAM));
static VOID    local DoBrowseClasses	Args((HWND, UINT, WPARAM, LPARAM));
static VOID    local DoBrowseTycons	Args((HWND, UINT, WPARAM, LPARAM));
static VOID    local DoBrowseNames	Args((HWND, UINT, WPARAM, LPARAM));
static INT     local DoCreate 	   	Args((HWND, UINT, WPARAM, LPARAM));
static VOID    local DoDestroy 	   	Args((HWND, UINT, WPARAM, LPARAM));
static VOID    local DoGetMinMaxInfo    Args((HWND, UINT, WPARAM, LPARAM));
static VOID    local DoLButtonDown 	Args((HWND, UINT, WPARAM, LPARAM));
static VOID    local DoLButtonUp 	Args((HWND, UINT, WPARAM, LPARAM));
static VOID    local DoMenuSelect 	Args((HWND, UINT, WPARAM, LPARAM));
static VOID    local DoMouseMove 	Args((HWND, UINT, WPARAM, LPARAM));
static VOID    local DoOptions 	   	Args((Void));
static VOID    local DoPaint		Args((HWND, UINT, WPARAM, LPARAM));
static VOID    local DoSize		Args((HWND, UINT, WPARAM, LPARAM));
static VOID    local DoInitMenu		Args((HWND, UINT, WPARAM, LPARAM));
static void    local DrawToolBar	Args((HWND));
static CHAR*   local GetaFileName	Args((HWND, UINT));
static VOID    local Hierarchy	   	Args((VOID));
static BOOL    local InitApplication    Args((VOID));
static BOOL    local InitInstance 	Args((LPSTR, INT));
static void    local LoadProfileOptions Args((VOID));
static void    local PutBitmap 	   	Args((HBITMAP, HWND, INT, INT, DWORD));
static void    local PutHelpLine	Args((HWND, LPCSTR));
static void    local SaveProfileOptions Args((VOID));

/* --------------------------------------------------------------------------
 * Nonlocal function prototypes:
 * ------------------------------------------------------------------------*/

INT APIENTRY WinMain	Args((HANDLE, HANDLE, LPSTR, INT));

/* --------------------------------------------------------------------------
 * CALLBACK functions prototypes:
 * ------------------------------------------------------------------------*/

BOOL    FAR APIENTRY _export AboutDlgProc	  (HWND, UINT, WPARAM, LPARAM);
BOOL    FAR APIENTRY _export BrowseClassesDlgProc (HWND, UINT, WPARAM, LPARAM);
BOOL    FAR APIENTRY _export BrowseTyconsDlgProc  (HWND, UINT, WPARAM, LPARAM);
BOOL    FAR APIENTRY _export BrowseNamesDlgProc   (HWND, UINT, WPARAM, LPARAM);
BOOL    FAR APIENTRY _export ScriptManDlgProc	  (HWND, UINT, WPARAM, LPARAM);

LRESULT FAR APIENTRY _export MainWndProc 	  (HWND, UINT, WPARAM, LPARAM);
extern LRESULT FAR APIENTRY _export ClassesWndProc(HWND, UINT, WPARAM, LPARAM);
BOOL    FAR APIENTRY _export OptionsDlgProc	  (HWND, UINT, WPARAM, LPARAM);

#define 	ON	TRUE
#define 	OFF	FALSE

#define ROWS		25
#define COLS 		80		/* Default Windows size in chars   */

/* --------------------------------------------------------------------------
 * Local Variables:
 * ------------------------------------------------------------------------*/

	HANDLE	hThisInstance;          /* Windows instance of application */
	HWND	hWndMain=NULL;		/* Main Window handle              */
	HWND	hWndText=NULL;		/* Text Window handle   	   */
	HWND	hWndClasses=NULL;	/* Class Hierarchy Window handle   */

static	HANDLE  hAccelTable;		/* Accelerators table		   */
static 	HFONT   theFont=NULL;		/* Text Font to use in text window */
static 	HFONT   theErrorFont=NULL;   	/* Text Font for error messagess   */
static 	HCURSOR NormalCursor=NULL;   	/* Normal mouse cursor		   */
static 	HCURSOR HandCursor=NULL; 	/* Tool Bar mouse cursor	   */
static 	HCURSOR GarbageCursor=NULL;	/* GC mouse cursor		   */
static 	HCURSOR SaveCursor;             /* Used to save and restore cursor */
static 	INT 	WindowWidth=0,
		WindowHeight=0;		/* Size of Main Window		   */
static  INT	NMenus;			/* Number of items in the menu	   */
static	INT	ScreenRows = 0,
		ScreenCols = 0;         /* Text Window size in chars	   */

#if BCWIN16
String appName = "Hugs for Windows";	/* Text for application name	   */
#else
String appName = "Hugs for Win 32";	/* Text for application name	   */
#endif

#define MAXBUF 300
CHAR buffer[MAXBUF];

/* --------------------------------------------------------------------------
 * Lower status line implementation:
 * ------------------------------------------------------------------------*/

RECT 	RectStatus;		/* Status Line position in main window     */
HBITMAP bmpStatusBar=NULL;	/* Status line Bitmap           	   */
#define STATUS_HEIGHT  19       /* Status line height           	   */
#define EDIT_LEFT     -120	/* Horizontal offset of "Editor:"          */

/* Draws the Status line on screen */
static VOID local DrawStatusLine(HWND hWnd) {
    HDC     hDC;
    RECT    rc;
    LOGFONT lf;
    HFONT   hSmall, hOld;
    INT     OldBkMode;
    CHAR    szMsg1[300];
    CHAR    drive[MAXDRIVE];
    CHAR    dir[MAXDIR];
    CHAR    file[MAXFILE];
    CHAR    ext[MAXEXT];

    PutBitmap(bmpStatusBar,hWnd,RectStatus.left,RectStatus.top,SRCCOPY);

    hDC = GetDC(hWnd);

    /* Get a small font */
    hOld = (HFONT) GetStockObject(ANSI_VAR_FONT);
    if (hOld) {
	GetObject(hOld,sizeof(LOGFONT),(LPSTR)&lf);
	lf.lfHeight = -1;
	lf.lfWidth  = 0;
	lf.lfWeight = FW_NORMAL;
	hSmall      = CreateFontIndirect(&lf);
    }
    else
	hSmall = 0;

    hOld      = hSmall ? ((HFONT) SelectObject(hDC, hSmall)) : 0;
    OldBkMode = SetBkMode(hDC, TRANSPARENT);
    rc.left   = RectStatus.right+EDIT_LEFT+12;
    rc.right  = RectStatus.right-7;
    rc.top    = RectStatus.top+4;
    rc.bottom = RectStatus.bottom-2;

    if (lastEdit) {
	fnsplit(lastEdit,drive,dir,file,ext);
	wsprintf(szMsg1,"Edit: %s%s%s",drive,file,ext);
    }
    else
	wsprintf(szMsg1,"Edit: %s","None");

    DrawText(hDC,(LPCSTR)szMsg1,strlen((CHAR *)szMsg1),&rc,DT_LEFT);
    SetBkMode(hDC,OldBkMode);

    if (hOld)
	SelectObject(hDC, hOld);
    if (hSmall)
	DeleteObject(hSmall);

    ReleaseDC(hWnd, hDC);
}

static HBITMAP local CreateStatusBar(HWND hWnd) {
    HBITMAP hbmpOldSrc, hbmpOldDest;	/* Create a bitmap for the status   */
    HBITMAP hbmpNew, hbmpSrc;		/* line, according to window size   */
    HDC     hdcSrc, hdcDest, hDC;
    BITMAP  bmp;

    hDC     = GetDC(hWnd);
    hdcSrc  = CreateCompatibleDC(hDC);
    hdcDest = CreateCompatibleDC(hDC);
    ReleaseDC(hWnd,hDC);

    hbmpSrc    = LoadBitmap(hThisInstance,"STATUSBAR");
    GetObject(hbmpSrc,sizeof(BITMAP),&bmp);
    hbmpOldSrc = SelectObject(hdcSrc,hbmpSrc);
    hbmpNew    = CreateCompatibleBitmap(hdcSrc,
					RectStatus.right-RectStatus.left+1,
					RectStatus.bottom-RectStatus.top+1);
    hbmpOldDest = SelectObject(hdcDest, hbmpNew);
    StretchBlt(hdcDest,0,0,RectStatus.right-RectStatus.left+1,
			   RectStatus.bottom-RectStatus.top+1,
	       hdcSrc,0,0, bmp.bmWidth, bmp.bmHeight,
	       SRCCOPY);
    SelectObject(hdcSrc,hbmpOldSrc);
    DeleteObject(hbmpSrc);

    hbmpSrc = LoadBitmap(hThisInstance,"STATUSBARSEP");
    GetObject(hbmpSrc,sizeof(BITMAP),&bmp);
    hbmpOldSrc = SelectObject(hdcSrc,hbmpSrc);
    StretchBlt(hdcDest,RectStatus.right+EDIT_LEFT,0, bmp.bmWidth,
			    RectStatus.bottom-RectStatus.top+1,
	       hdcSrc,0,0, bmp.bmWidth, bmp.bmHeight,
	       SRCCOPY);
    SelectObject(hdcSrc, hbmpOldSrc);
    DeleteObject(hbmpSrc);

    SelectObject(hdcDest, hbmpOldDest);
    DeleteDC(hdcDest);
    DeleteDC(hdcSrc);
    return hbmpNew;
}

static void local PutHelpLine(HWND hWnd, LPCSTR szMsg) {
    INT     iLen = strlen(szMsg);	/* Print message szMsg on the status*/
    HDC     hDC  = GetDC(hWnd);		/* line of window hWnd		    */
    RECT    rc;
    LOGFONT lf;
    HFONT   hSmall, hOld;
    INT     OldBkMode;

    if (hOld=(HFONT)GetStockObject(ANSI_VAR_FONT)) {	/* get small font  */
	GetObject(hOld,sizeof(LOGFONT),(LPSTR)&lf);
	lf.lfHeight = -1;
	lf.lfWidth  = 0;
	lf.lfWeight = FW_NORMAL;
	hSmall      = CreateFontIndirect(&lf);
    }
    else
	hSmall = 0;

    hOld      = hSmall ? (HFONT)(SelectObject(hDC,hSmall)) : 0;
    OldBkMode = SetBkMode (hDC, TRANSPARENT);
    rc.left   = RectStatus.left+5;
    rc.right  = RectStatus.right+EDIT_LEFT-2;
    rc.top    = RectStatus.top+4;
    rc.bottom = RectStatus.bottom-2;
    FillRect(hDC,&rc,GetStockObject(LTGRAY_BRUSH));
    DrawText(hDC,szMsg,iLen,&rc,DT_LEFT);
    SetBkMode(hDC,OldBkMode);

    if (hOld)
	SelectObject(hDC, hOld);
    if (hSmall)
	DeleteObject(hSmall);
    ReleaseDC(hWnd, hDC);

    return;
}

/* --------------------------------------------------------------------------
 * Toolbar implementation:
 * ------------------------------------------------------------------------*/

static RECT RectTools;		/* Tool bar position                        */
static INT  Button_Height;	/* Size of one button                      */
static INT  Button_Width;
static INT  nButtons;		/* Number of buttons in the tool bar	   */
static INT  buttonPushed = -1;
static Bool buttonDown   = FALSE;

#define TOOLS_WIDTH	(Button_Width+9)               /* Tool bar width   */
#define TOOLS_BK_COLOR  RGB(192,192,192)               /* Tool bar color   */

typedef struct tagButton {	/* One button on the tool bar:		   */
    WPARAM Msg;			/* Command to execute when pushed	   */
    BOOL   IsEnabled;		/* Is the button enabled?		   */
    LPCSTR Bitmap;		/* Bitmap name for the button		   */
    UINT   IdHelpLine;		/* Help to get when pointed		   */
} hButton;

static hButton theButtons[] = {	/* The tool bar buttons			   */
    {ID_HELPINDEX,       TRUE,  "HELPBUTTON",      ID_HELPINDEX},
    {ID_SCRIPTMAN,    	 TRUE,  "SCRIPTMANBUTTON", ID_SCRIPTMAN},
    {ID_RELOAD,          TRUE,  "RELOADBUTTON",    ID_RELOAD},
    {ID_COPY,        	 TRUE,  "COPYBUTTON",      ID_COPY},
    {ID_CUT,         	 TRUE,  "CUTBUTTON",	   ID_CUT},
    {ID_PASTE,       	 TRUE,  "PASTEBUTTON",     ID_PASTE},
    {ID_CLEAR,       	 TRUE,  "DELETEBUTTON",    ID_CLEAR},
    {ID_RUN,	       	 TRUE,  "RUNBUTTON",	   ID_RUN},
    {ID_STOP,       	 TRUE,  "STOPBUTTON",      ID_STOP},
    {ID_GOEDIT,       	 TRUE,  "EDITBUTTON",      ID_GOEDIT},
    {ID_SETOPTIONS,      TRUE,  "OPTIONSBUTTON",   ID_SETOPTIONS},
    {ID_BROWSEHIERARCHY, TRUE,  "HIERARCHYBUTTON", ID_BROWSEHIERARCHY},
    {ID_EXIT,        	 TRUE,  "EXITBUTTON",      ID_EXIT},
    {0, 0, 0, 0}
};

static VOID local DrawToolBar(HWND hWnd) { /* Draw tool bar on the screen  */
    HDC     hDC = GetDC(hWnd);
    HBRUSH  hBrush, hOldBrush;
    HPEN    hPen, hOldPen;
    HBITMAP hBitmap1, hBitmap2;
    INT     i;

    hBrush    = CreateSolidBrush(TOOLS_BK_COLOR);
    hOldBrush = SelectObject(hDC,hBrush);
    FillRect(hDC,&RectTools,hBrush);
    SelectObject(hDC,hOldBrush);
    DeleteObject(hBrush);

    hPen    = CreatePen(PS_SOLID,1,RGB(0,0,0));
    hOldPen = SelectObject(hDC,hPen);
    MoveToEx(hDC,RectTools.right-1,RectTools.top,NULL);
    LineTo(hDC,RectTools.right-1,RectTools.bottom);
    SelectObject(hDC, hOldPen);
    DeleteObject(hPen);

    hBitmap1 = LoadBitmap(hThisInstance,"BUTTON");
    for (i=0; theButtons[i].Msg; i++) {
	PutBitmap(hBitmap1,hWnd,RectTools.left+4,
				RectTools.top+(i*Button_Height)+1,
				SRCCOPY);
	hBitmap2 = LoadBitmap(hThisInstance, theButtons[i].Bitmap);
	PutBitmap(hBitmap2,hWnd,RectTools.left+7,
				RectTools.top+(i*Button_Height)+2+1,
				SRCCOPY);
	DeleteObject(hBitmap2);
	if (!theButtons[i].IsEnabled) {
	    hBitmap2 = LoadBitmap(hThisInstance,"DISABLED");
	    PutBitmap(hBitmap2,hWnd,RectTools.left+7,
				    RectTools.top+(i*Button_Height)+2+1,
				    SRCAND);
	    DeleteObject(hBitmap2);
	}
    }
    DeleteObject(hBitmap1);
    ReleaseDC(hWnd, hDC);
}

static void pushButton(HWND hWnd) {
    if (!buttonDown) {
	HBITMAP hBitmap = LoadBitmap(hThisInstance,"PUSHEDBUTTON");
	PutBitmap(hBitmap,hWnd,RectTools.left+4,
			       RectTools.top+(buttonPushed*Button_Height)+1,
			       SRCCOPY);
	DeleteObject(hBitmap);
	hBitmap = LoadBitmap(hThisInstance, theButtons[buttonPushed].Bitmap);
	PutBitmap(hBitmap,hWnd,RectTools.left+8,
			       RectTools.top+(buttonPushed*Button_Height)+3+1,
			       SRCCOPY);
	DeleteObject(hBitmap);
	buttonDown = TRUE;
    }
}

static void unpushButton(HWND hWnd) {
    if (buttonDown) {
	HBITMAP hBitmap = LoadBitmap(hThisInstance,"BUTTON");
	PutBitmap(hBitmap,hWnd,RectTools.left+4,
			       RectTools.top+(buttonPushed*Button_Height)+1,
			       SRCCOPY);
	DeleteObject(hBitmap);
	hBitmap = LoadBitmap(hThisInstance,theButtons[buttonPushed].Bitmap);
	PutBitmap(hBitmap,hWnd,RectTools.left+7,
			       RectTools.top+(buttonPushed*Button_Height)+2+1,
			       SRCCOPY);
	DeleteObject(hBitmap);
	buttonDown = FALSE;
    }
}

/* --------------------------------------------------------------------------
 * Other MS-DOS functions emulation:
 * ------------------------------------------------------------------------*/

INT WinSystem(CHAR *s) {
    ERRTEXT "\nCannot execute: %s\n", s
    EEND;
    return 0;
}

/* --------------------------------------------------------------------------
 * Responses to messages:
 * ------------------------------------------------------------------------*/

static HWND ahwndSubMenus[16];

/* Browse Classes ... */
#pragma argsused
static VOID local DoBrowseClasses(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  DLGPROC lpProcBrowse;

  lpProcBrowse = (DLGPROC)MakeProcInstance((FARPROC)BrowseClassesDlgProc, hThisInstance);
  DialogBox(hThisInstance,MAKEINTRESOURCE(BROWSECLASSESDLGBOX),GetFocus(),lpProcBrowse);
  FreeProcInstance(lpProcBrowse);
}

/* Browse Type Constructors ... */
#pragma argsused
static VOID local DoBrowseTycons(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  DLGPROC lpProcBrowse;

  lpProcBrowse = (DLGPROC)MakeProcInstance((FARPROC)BrowseTyconsDlgProc, hThisInstance);
  DialogBox(hThisInstance, MAKEINTRESOURCE(BROWSETYCONSDLGBOX), GetFocus(), lpProcBrowse);
  FreeProcInstance(lpProcBrowse);
}

/* Browse Names ... */
#pragma argsused
static VOID local DoBrowseNames(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  DLGPROC lpProcBrowse;

  lpProcBrowse = (DLGPROC)MakeProcInstance((FARPROC)BrowseNamesDlgProc, hThisInstance);
  DialogBox(hThisInstance, MAKEINTRESOURCE(BROWSENAMESDLGBOX), GetFocus(), lpProcBrowse);
  FreeProcInstance(lpProcBrowse);
}

/* Script Manager */
#pragma argsused
static VOID local ScriptManager(HWND hWnd, UINT msg,
				WPARAM wParam, LPARAM lParam) {
  DLGPROC dlgProc = (DLGPROC)MakeProcInstance((FARPROC)ScriptManDlgProc,
					      hThisInstance);
  DialogBox(hThisInstance,
	    MAKEINTRESOURCE(SCRIPTMANDLGBOX),
	    GetFocus(),
	    dlgProc);
  FreeProcInstance(dlgProc);
}

/* Response to Menu Commands */
#pragma argsused
static VOID local DoCommand(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
    switch (wParam) {
	/* Load script from disk (name of the script is currently selected */
	case ID_OPENSELECTED: {
	    String File = GetSelectedText(hWndText);
	    input(BREAK);
            WinPuts(hWndText, "\n");
            setLastEdit(File, 0);
            if (projectLoaded) {
                if (MessageBox(GetFocus(),
			       "Close current loaded project?",
			       appName,
			       MB_ICONQUESTION|MB_YESNO)!=IDYES) {
		    whatFiles();
                    longjmp(catch_error, 1);
                }
                clearProject();
		forgetScriptsFrom(1);
	    }
	    addScriptName(File,TRUE);
	    readScripts(1);
	    longjmp(catch_error, 1);
	    break;
	}

	case ID_EXIT:			/* Exit hugs interpreter 	   */
	    SendMessage(hWnd, WM_CLOSE, 0, 0L);
	    break;

	case ID_COPY:			/* Clipboard copy		   */
	    SendMessage(hWndText,WM_COPY,0,0L);
	    break;

	case ID_PASTE:			/* Clipboard paste		   */
	    SendMessage(hWndText, WM_PASTE, 0, 0L);
	    break;

	case ID_CUT:			/* Clipboard cut		   */
	    SendMessage(hWndText, WM_CUT, 0, 0L);
	    break;

	case ID_CLEAR:			/* Clipboard clear		   */
	    SendMessage(hWndText, WM_CLEAR, 0, 0L);
	    break;

	case ID_GOEDIT:			/* Open text editor		   */
	    runEditor();
	    break;

	case ID_EDITSELECTED:		/* Open editor for selected text   */
	    setLastEdit((String)GetSelectedText(hWndText), 0);
	    runEditor();
	    break;

        case ID_FIND: {			/* Find defn of selected text	   */
	    String nm = GetSelectedText(hWndText);
	    input(BREAK);
            WinPuts(hWndText, "\n");
	    SendMessage(hWndText, WM_CHAR, (WPARAM) ':',       0L);
            SendMessage(hWndText, WM_CHAR, (WPARAM) 'f',       0L);
            SendMessage(hWndText, WM_CHAR, (WPARAM) ' ',       0L);
	    for (;*nm;nm++)
		SendMessage(hWndText, WM_CHAR, (WPARAM) *nm,     0L);
	    SendMessage(hWndText, WM_CHAR, (WPARAM) VK_RETURN, 0L);
	    longjmp(catch_error, 1);
	    break;
	}

        case ID_TYPE: {			/* Show type of selected text	   */
	    String nm = GetSelectedText(hWndText);
	    input(BREAK);
            WinPuts(hWndText, "\n");
	    SendMessage(hWndText, WM_CHAR, (WPARAM) ':',       0L);
            SendMessage(hWndText, WM_CHAR, (WPARAM) 't',       0L);
            SendMessage(hWndText, WM_CHAR, (WPARAM) ' ',       0L);
            for (;*nm;nm++)
		SendMessage(hWndText, WM_CHAR, (WPARAM) *nm,     0L);
            SendMessage(hWndText, WM_CHAR, (WPARAM) VK_RETURN, 0L);
	    longjmp(catch_error, 1);
	    break;
	}

        case ID_INFO: {			/* Show info on selected text	   */
	    String nm = GetSelectedText(hWndText);
	    input(BREAK);
            WinPuts(hWndText, "\n");
            SendMessage(hWndText, WM_CHAR, (WPARAM) ':',       0L);
            SendMessage(hWndText, WM_CHAR, (WPARAM) 'i',       0L);
            SendMessage(hWndText, WM_CHAR, (WPARAM) ' ',       0L);
            for (;*nm;nm++)
                SendMessage(hWndText, WM_CHAR, (WPARAM) *nm,     0L);
            SendMessage(hWndText, WM_CHAR, (WPARAM) VK_RETURN, 0L);
	    longjmp(catch_error, 1);
            break;
	}

        case ID_EVAL: {			/* Eval selected text		  */
	    String nm = GetSelectedText(hWndText);
	    input(BREAK);
            WinPuts(hWndText, "\n");
	    for (;*nm;nm++)
                SendMessage(hWndText, WM_CHAR, (WPARAM) *nm,     0L);
            SendMessage(hWndText, WM_CHAR, (WPARAM) VK_RETURN, 0L);
            longjmp(catch_error, 1);
            break;
	}

	case ID_STOP:			/* Stop program execution	   */
	    breakHandler();
	    break;

	case ID_RUN:			/* Evaluate main expression	   */
	    input(BREAK);
	    WinPuts(hWndText, "\n");
	    readScripts(1);
	    SendMessage(hWndText, WM_CHAR, (WPARAM) 'm',       0L);
	    SendMessage(hWndText, WM_CHAR, (WPARAM) 'a',       0L);
	    SendMessage(hWndText, WM_CHAR, (WPARAM) 'i',       0L);
	    SendMessage(hWndText, WM_CHAR, (WPARAM) 'n',       0L);
	    SendMessage(hWndText, WM_CHAR, (WPARAM) VK_RETURN, 0L);
	    longjmp(catch_error, 1);
	    break;

	case ID_SETOPTIONS:		/* Set interpreter options	   */
	    DoOptions();
	    SaveProfileOptions();
	    break;

	case ID_LOADOPTIONS:		/* Load options from Hugs.Ini	   */
	    LoadProfileOptions();
	    break;

	case ID_SAVEOPTIONS:		/* Save options to Hugs.Ini	   */
	    SaveProfileOptions();
	    break;

	case ID_EDITOPTIONS: {		/* Open Hugs.Ini file in editor	   */
	    CHAR buffer[FILENAME_MAX+1];
	    GetWindowsDirectory(buffer, FILENAME_MAX);
	    strncat(buffer,"\\Hugs.ini",FILENAME_MAX - strlen(buffer));
	    startEdit(0,buffer);
	    break;
	}

	case ID_RELOAD:			/* Reload script files loaded	   */
	    input(BREAK);
	    WinPuts(hWndText, "\n");
	    readScripts(1);
	    longjmp(catch_error, 1);
	    break;

	case ID_BROWSENAMES:		/* Browse names			   */
	    DoBrowseNames(hWnd, msg, wParam, lParam);
	    break;

	case ID_BROWSETYCONS:		/* Browse type constructors	   */
	    DoBrowseTycons(hWnd, msg, wParam, lParam);
	    break;

	case ID_BROWSECLASSES:		/* Browse classes		   */
	    DoBrowseClasses(hWnd, msg, wParam, lParam);
	    break;

	case ID_BROWSEHIERARCHY:	/* Show class hierarchy		   */
	    Hierarchy();
	    break;

	case ID_SCRIPTMAN:		/* Enter Script Manager		   */
	    ScriptManager(hWnd, msg, wParam, lParam);
	    break;

       case ID_HELPUSE:			/* help about using windows help   */
	    WinHelp(hWnd, "Winhelp.Hlp", HELP_CONTENTS, 0L);
	    break;

	case ID_HELPINDEX:		/* Show help about Hugs		   */
	    GetPrivateProfileString("Help","Path","\\hugs\\docs\\hugs.hlp",
				    buffer, MAXBUF, "Hugs.ini");
	    WinHelp(hWnd, buffer, HELP_CONTENTS, 0L);
	    break;

	case ID_ABOUT:			/* Show Hugs copyright		   */
	    DoAbout();
	    break;
    }
}

/* Respond to Windows WM_CREATE message */
#pragma argsused
static INT local DoCreate(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  BITMAP     bm;
  HBITMAP    hBitmap;
  INT        i;

  /* Create a font to use in text window */
  theFont = CreateFont(15, 7, 0, 0,
			 FW_NORMAL,
			 FALSE,
			 FALSE,
			 FALSE,
			 ANSI_CHARSET,
			 OUT_TT_PRECIS,
			 CLIP_TT_ALWAYS,
			 DEFAULT_QUALITY,
			 FIXED_PITCH | FF_MODERN,
			 "Courier New");

  if (!theFont) {
    MessageBox(GetFocus(), "Out of memory creating text font", appName,
	       MB_ICONHAND | MB_SYSTEMMODAL | MB_OK);
    DoDestroy(hWnd, msg, wParam, lParam);
    return -1;
  }

  /* Create a font for error messages */
  theErrorFont = CreateFont( 15, 7, 0, 0,
			     FW_NORMAL,
			     TRUE,  /* Italic */
			     FALSE,
			     FALSE,
			     ANSI_CHARSET,
			     OUT_TT_PRECIS,
			     CLIP_TT_ALWAYS,
			     DEFAULT_QUALITY,
			     FIXED_PITCH | FF_MODERN,
			     "Courier New");

  if (!theErrorFont) {
    MessageBox(GetFocus(), "Out of memory creating error font", appName,
	       MB_ICONHAND | MB_SYSTEMMODAL | MB_OK);
    DoDestroy(hWnd, msg, wParam, lParam);
    return -1;
  }

  /* Load Windows cursor to use */
  NormalCursor  = LoadCursor(NULL, IDC_ARROW);
  HandCursor    = LoadCursor(hThisInstance, "HANDCURSOR");
  GarbageCursor = LoadCursor(hThisInstance, "GARBAGECURSOR");

  if (!(NormalCursor && HandCursor && GarbageCursor)){
    MessageBox(GetFocus(), "Out of memory loading cursors", appName,
	       MB_ICONHAND | MB_SYSTEMMODAL | MB_OK);
    DoDestroy(hWnd, msg, wParam, lParam);
    return -1;
  }

  /* Get buttons size */
  hBitmap = LoadBitmap(hThisInstance, "BUTTON");
  GetObject(hBitmap, sizeof(BITMAP), &bm);
  Button_Width  = bm.bmWidth+2;
  Button_Height = bm.bmHeight+2;
  DeleteObject(hBitmap);

  /* Get Handles to all submenus (needed for status line messages) */
  NMenus = GetMenuItemCount(GetMenu(hWnd));
  for (i=0; i<NMenus; ahwndSubMenus[i]=GetSubMenu(GetMenu(hWnd),i), i++);

  /* Get number of buttons in tool bar */
  for(nButtons=0; theButtons[nButtons].Msg; nButtons++);

  /* Use 3-D look in dialog boxes */
  Ctl3dRegister(hThisInstance);
  Ctl3dAutoSubclass(hThisInstance);

  return 0;
}

/* Respond to Windows WM_DESTROY message */
#pragma argsused
static VOID local DoDestroy(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam) {
    /* Close Help Windows, if opened */
    WinHelp(hWnd, "WINHELP.HLP", HELP_QUIT, 0L);
    WinHelp(hWnd, "HUGS.HLP",   HELP_QUIT, 0L);

    /* Free space used for Fonts and cursors */
    if (theFont)       DeleteObject(theFont);
    if (theErrorFont)  DeleteObject(theErrorFont);
    if (NormalCursor)  DestroyCursor(NormalCursor);
    if (HandCursor)    DestroyCursor(HandCursor);
    if (GarbageCursor) DestroyCursor(GarbageCursor);

    Ctl3dUnregister(hThisInstance);	/* No more 3-D look */
}

/* Respond to Windows WM_GETMINMAXINFO message */
#pragma argsused
static VOID local DoGetMinMaxInfo(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    MINMAXINFO FAR* lpmmi;
    RECT	    rWind;

    GetWindowRect(hWnd, &rWind);

    lpmmi 		    = (MINMAXINFO FAR*) lParam;
    lpmmi->ptMaxPosition.x  = rWind.left;
    lpmmi->ptMaxPosition.y  = rWind.top;
    lpmmi->ptMaxTrackSize.x = WindowWidth;
    lpmmi->ptMaxTrackSize.y = WindowHeight;
    lpmmi->ptMaxSize.x      = WindowWidth;
    lpmmi->ptMaxSize.y      = WindowHeight;
}

/* Respond to Windows WM_LBUTTONDOWN message */
#pragma argsused
static VOID local DoLButtonDown(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
   INT     xPos = LOWORD(lParam);
   INT     yPos = HIWORD(lParam);
   INT     nButton;

   /* Check if a button on the Tool bar is pushed */
   if (xPos < RectTools.right &&
       yPos && yPos < RectTools.top + 1 + nButtons*(Button_Height)){
     nButton = (yPos) / (Button_Height);
     if (theButtons[nButton].IsEnabled) {
      buttonPushed = nButton;
      SetCapture(hWnd);
      pushButton(hWnd);
     }
   }
}

/* Respond to Windows WM_LBUTTONUP message */
#pragma argsused
static VOID local DoLButtonUp(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    if (buttonPushed != -1) {
      if (buttonDown) {
	SetCursor(NormalCursor);
	PostMessage(hWnd, WM_COMMAND, theButtons[buttonPushed].Msg, 0L);
	PutHelpLine(hWnd, "");
	unpushButton(hWnd);
      }
      buttonPushed = (-1);
      ReleaseCapture();
    }
}

/* Used to show a message on the help line depending on the menu option selected */
#pragma argsused
static VOID local DoMenuSelect(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  WORD idMenu;
  CHAR szMsg[80];
#if BCWIN32
  UINT Item=(UINT)LOWORD(wParam), Flags=(UINT)HIWORD(wParam);
#else
  UINT Item=wParam, Flags=lParam;
#endif

  strcpy(szMsg, "");

  if (Flags & MF_POPUP){ /* It is a submenu (like Files) */
#if BCWIN32
    for (idMenu=0; idMenu < 16 && GetSubMenu((HMENU)lParam,Item)!=(HMENU)ahwndSubMenus[idMenu]; idMenu++);
#else
    for (idMenu=0; idMenu < 16 && Item!=(UINT)ahwndSubMenus[idMenu]; idMenu++);
#endif

   if (!ahwndSubMenus[idMenu] || idMenu == 16) {
    PutHelpLine(hWnd, "");
    goto end;
   }
   idMenu++; /* offsets begin with 1 */
  }
  else
   idMenu = Item;  /* Item is the curent menu item identifier */

  if ((Flags == 0xffff) ||
     ((LoadString(hThisInstance, idMenu, (LPSTR) szMsg, sizeof(szMsg))) != NULL )){

   if (Flags == 0xffff)
    PutHelpLine(hWnd, ""); /* The menu has been closed */
   else
    PutHelpLine(hWnd, szMsg);
  }
  end:
}

/* Respond to Windows WM_MOUSEMOVE message */
#pragma argsused
static VOID local DoMouseMove(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
   INT  xPos = LOWORD(lParam);
   INT  yPos = HIWORD(lParam);
   INT  nButton;
   CHAR szMsg[256];

   static INT lastHelp = -1;

   /* Show help if cursor on a button of the tool bar */
   if (xPos < RectTools.left+Button_Width+4 && yPos && yPos < RectTools.top + 1 + nButtons*(Button_Height)){
     SetCursor(HandCursor);
     nButton = yPos / (Button_Height);
     if (buttonPushed!=(-1))
       if (buttonPushed==nButton)
	 pushButton(hWnd);
       else
	 unpushButton(hWnd);
     if (lastHelp != nButton) {
      lastHelp = nButton;
      LoadString(hThisInstance,
		 theButtons[nButton].IdHelpLine,
		 (LPSTR) szMsg,
		 sizeof(szMsg));
      PutHelpLine(hWnd, (LPSTR)szMsg);
     }
   }
   else {
     unpushButton(hWnd);
     SetCursor(NormalCursor);
     if (lastHelp != -1) {
      lastHelp = -1;
      PutHelpLine(hWnd, "");
     }
   }
}

/* Respond to Windows WM_PAINT message */
#pragma argsused
static VOID local DoPaint(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
  PAINTSTRUCT ps;
  RECT RectIntersect;

  BeginPaint(hWnd, &ps);
  /* Paint Tool bar if needed */
  if (IntersectRect(&RectIntersect, &RectTools, &ps.rcPaint))
   DrawToolBar(hWnd);
  /* Paint Status line if needed */
  if (IntersectRect(&RectIntersect, &RectStatus, &ps.rcPaint))
   DrawStatusLine(hWnd);
  EndPaint(hWnd, &ps);
}

/* Respond to Windows WM_SIZE message */
#pragma argsused
static VOID local DoSize(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
  RECT rClient;

  GetClientRect(hWnd, (LPRECT) &rClient);

  /* Set lower status line position */
  RectStatus.top    = rClient.bottom - STATUS_HEIGHT;
  RectStatus.bottom = rClient.bottom;
  RectStatus.left   = rClient.left;
  RectStatus.right  = rClient.right;

  /* Set Tool Bar position */
  RectTools.top = rClient.top;
  RectTools.bottom = rClient.bottom - STATUS_HEIGHT;
  RectTools.left   = rClient.left;
  RectTools.right  = rClient.left+TOOLS_WIDTH-1;

  /* Resize status line */
  if (bmpStatusBar)
    DeleteObject(bmpStatusBar);

  bmpStatusBar = CreateStatusBar(hWnd);

  #define H_INDENT     0    /* Separation between tool bar and text window */
  #define V_INDENT     0    /* Separation between main menu and text window */

  /* Resize child text window size */
  MoveWindow(hWndText, TOOLS_WIDTH-1+H_INDENT, V_INDENT,
    rClient.right-RectTools.right-H_INDENT*2, RectStatus.top-V_INDENT*2, TRUE);
}

#pragma argsused
static VOID local DoInitMenu (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
  BOOL CanCopy  = (BOOL) SendMessage (hWndText, WM_CANCOPY,  0, 0L);
  BOOL CanCut   = (BOOL) SendMessage (hWndText, WM_CANCUT,   0, 0L);
  BOOL CanPaste = (BOOL) SendMessage (hWndText, WM_CANPASTE, 0, 0L);
  BOOL CanClear = (BOOL) SendMessage (hWndText, WM_CANCLEAR, 0, 0L);
  EnableMenuItem((HMENU)wParam, ID_COPY,  (CanCopy  ? MF_ENABLED : MF_GRAYED)|MF_BYCOMMAND);
  EnableMenuItem((HMENU)wParam, ID_PASTE, (CanPaste ? MF_ENABLED : MF_GRAYED)|MF_BYCOMMAND);
  EnableMenuItem((HMENU)wParam, ID_CUT,   (CanCut   ? MF_ENABLED : MF_GRAYED)|MF_BYCOMMAND);
  EnableMenuItem((HMENU)wParam, ID_CLEAR, (CanClear ? MF_ENABLED : MF_GRAYED)|MF_BYCOMMAND);
  EnableMenuItem((HMENU)wParam, ID_EVAL,  (CanCopy  ? MF_ENABLED : MF_GRAYED)|MF_BYCOMMAND);
  EnableMenuItem((HMENU)wParam, ID_TYPE,  (CanCopy  ? MF_ENABLED : MF_GRAYED)|MF_BYCOMMAND);
  EnableMenuItem((HMENU)wParam, ID_INFO,  (CanCopy  ? MF_ENABLED : MF_GRAYED)|MF_BYCOMMAND);
  EnableMenuItem((HMENU)wParam, ID_FIND,  (CanCopy  ? MF_ENABLED : MF_GRAYED)|MF_BYCOMMAND);
  EnableMenuItem((HMENU)wParam, ID_OPENSELECTED,  (CanCopy  ? MF_ENABLED : MF_GRAYED)|MF_BYCOMMAND);
  EnableMenuItem((HMENU)wParam, ID_EDITSELECTED,  (CanCopy  ? MF_ENABLED : MF_GRAYED)|MF_BYCOMMAND);
}

/* --------------------------------------------------------------------------
 * Other functions for Windows GUI:
 * ------------------------------------------------------------------------*/

static HBITMAP local CreatehDC (HWND hWnd, HDC *hBmpDC, LPCSTR bitmapName, COLORREF Color) {
  INT       i, j;
  HDC       hDC;
  HBITMAP   hbmp, hOldBitmap;
  BITMAP    bm;

  /* Create Doc and NoDoc Bitmaps */
  hDC = GetDC (hWnd);

  hbmp = LoadBitmap(hThisInstance, bitmapName);
  GetObject(hbmp, sizeof(BITMAP), &bm);

  *hBmpDC = CreateCompatibleDC(hDC);
  hOldBitmap = SelectObject(*hBmpDC, hbmp);

  /* Set Background color for bitmap */
  for(i=0; i<bm.bmWidth; i++)
    for(j=0; j<bm.bmHeight; j++) {
      if (GetPixel(*hBmpDC, i, j) == RGB(0,128,128)) {
	 SetPixel(*hBmpDC, i, j, Color);
      }
    }

  ReleaseDC (hWnd, hDC);

  return hOldBitmap;
}

/* Draw a bitmap on a Window */
static void local PutBitmap(HBITMAP hBitmap, HWND hWnd, INT left, INT top, DWORD mode)
{
   HBITMAP hOldBitmap;
   BITMAP  bm;
   HDC     hDC, hDCMemory;

   hDC = GetDC(hWnd);

   GetObject(hBitmap, sizeof(BITMAP), &bm);
   hDCMemory = CreateCompatibleDC(hDC);
   hOldBitmap = SelectObject(hDCMemory, hBitmap);
   BitBlt(hDC, left, top, bm.bmWidth, bm.bmHeight, hDCMemory, 0, 0, mode);

   SelectObject(hDCMemory, hOldBitmap);
   DeleteDC(hDCMemory);

   ReleaseDC(hWnd, hDC);
}

/* Used to get a file name using a common dialog box. Returns a pointer to    */
/* a static string where the file name is, or NULL if cancel button is pushed */
/* Mask is the identificator of a string defined in Hugs.Rc where the type   */
/* of files and their masks are                                               */
static CHAR* local GetaFileName(HWND hWnd, UINT Mask) {
  static OPENFILENAME ofn;
  static CHAR szFileName[256];
  CHAR  szFile[256], szFileTitle[256];
  UINT  i, cbString;
  CHAR  chReplace;    /* Separator between diferent filters in szFilter */
  CHAR  szFilter[256];

  szFile[0] = '\0';

  if ((cbString = LoadString(hThisInstance, Mask,
       szFilter, sizeof(szFilter))) == 0) {
    /* Error */
    return NULL;
  }

  chReplace = szFilter[cbString - 1]; /* Get separator */

  /* Replace separator with NULL */
  for (i = 0; szFilter[i] != '\0'; i++) {
    if (szFilter[i] == chReplace)
       szFilter[i] = '\0';
  }

  memset(&ofn, 0, sizeof(OPENFILENAME));

  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hInstance = hThisInstance;
  ofn.hwndOwner = hWnd;
  ofn.lpstrFilter = szFilter;
  ofn.nFilterIndex = 1;
  ofn.lpstrFile= szFile;
  ofn.nMaxFile = sizeof(szFile);
  ofn.lpstrFileTitle = szFileTitle;
  ofn.nMaxFileTitle = sizeof(szFileTitle);
  ofn.lpTemplateName = MAKEINTRESOURCE(OPENFILEDLGBOX);
  ofn.Flags = OFN_HIDEREADONLY
/*          | OFN_PATHMUSTEXIST
	    | OFN_FILEMUSTEXIST */
	    | OFN_ENABLETEMPLATE;

  if (GetOpenFileName(&ofn)) {
    strcpy(szFileName, ofn.lpstrFile);
    strlwr(szFileName);
    return szFileName;
  }
  else
    return NULL;
}


/* Program entry point */
INT PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
		   LPSTR lpszCmdLine, INT nCmdShow) {
  INT	argc, i;
  CHAR  **argv, *p, *nextParam, saveChar;

  hThisInstance = hInstance;

  if (!hPrevInstance) {
   if (!InitApplication()) {
    MessageBeep(0);
#if BCWIN32
    {  DWORD Error;
       Error = GetLastError();
       sprintf (buffer, "Error number %u",Error);
       MessageBox (NULL, buffer, "Error", MB_OK);
    }
#endif
    return FALSE;
   }
  }

  /* Set Hugs interpreter options */
  LoadProfileOptions();

  if (!InitInstance(lpszCmdLine, nCmdShow))
   return FALSE;

  /* Allow other program to run */
  WinKbhit(hWndText);

  /* Get argc and argv */
  if (lpszCmdLine[0]) {
#if BCWIN16
   for(i=0, argc=1; lpszCmdLine[i]; i++)
#else
   for(i=0, argc=0; lpszCmdLine[i]; i++)
#endif
    if(lpszCmdLine[i] == ' ')
     argc++;
#if BCWIN16
   argc++;
#endif
  }
  else
   argc = 1;

  argv = malloc(argc*sizeof(CHAR *));

#if BCWIN16
  argv[0] = appName;
#endif

  if (lpszCmdLine[0]) {

   p=lpszCmdLine;

#if BCWIN16
   for(i=1; i<argc; i++) {
#else
   for(i=0; i<argc; i++) {
#endif
    for(nextParam=p; *p != ' ' && *p; p++)

    saveChar = *p;
    *p = '\0';

    argv[i] = strCopy(nextParam);
    *p = saveChar;

    p++;
    while (*p == ' ') p++;
   }
  }

  /* Call main */
  main(argc, argv);
  return 0;
}

/* Init application */
static BOOL local InitApplication(VOID) {
    WNDCLASS wc;
    wc.style	      = CS_VREDRAW | CS_HREDRAW;
    wc.lpfnWndProc    = MainWndProc;
    wc.cbClsExtra     = 0;
    wc.cbWndExtra     = 0;
    wc.hInstance      = hThisInstance;
    wc.hIcon	      = LoadIcon(hThisInstance, "Hugs");
    wc.hCursor	      = NULL; /* It must be NULL to use different ones */
    wc.hbrBackground  = GetStockObject(WHITE_BRUSH);
    wc.lpszMenuName   = "HugsMenu";
    wc.lpszClassName  = "HugsMainWindow";

    if (!RegisterClass(&wc))
	return FALSE;

    wc.style	      = CS_VREDRAW | CS_HREDRAW;
    wc.lpfnWndProc    = ClassesWndProc;
    wc.cbClsExtra     = 0;
    wc.cbWndExtra     = 0;
    wc.hInstance      = hThisInstance;
    wc.hIcon	      = NULL;
    wc.hCursor	      = NULL;
    wc.hbrBackground  = GetStockObject(LTGRAY_BRUSH);
    wc.lpszMenuName   = NULL;
    wc.lpszClassName  = "HugsClassesWindow";

    if (!RegisterClass(&wc))
	return FALSE;

    return RegisterTextClass(hThisInstance);
}

/* Init one instance of the application */
#pragma argsused
static BOOL local InitInstance(LPSTR lpCmdLine, INT nCmdShow) {
  RECT           rWind, rText;

  /* Create main window */
  hWndMain = CreateWindow("HugsMainWindow",
			  appName,
			  WS_OVERLAPPED  | WS_CAPTION | WS_SYSMENU |
			  WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_THICKFRAME,
			  CW_USEDEFAULT,
			  CW_USEDEFAULT,
			  CW_USEDEFAULT,
			  CW_USEDEFAULT,
			  (HWND) NULL,
			  (HMENU) NULL,
			  hThisInstance,
			  (LPSTR) NULL);

  if (!hWndMain)
   return FALSE;

  /* Load accelerators */
  hAccelTable = LoadAccelerators(hThisInstance, (LPSTR) "HugsAccelerators");

  hWndText = CreateTextWindow(hThisInstance,
			      hWndMain,
			      TOOLS_WIDTH+H_INDENT,
			      V_INDENT,
			      ScreenCols, ScreenRows,
			      theFont,
			      hAccelTable);
  if (!hWndText)
    return FALSE;

  /* Set main window size */
  GetWindowRect (hWndText, &rText);

  WindowWidth =  (rText.right-rText.left+1)+
		 GetSystemMetrics(SM_CXFRAME)*2+
		 GetSystemMetrics(SM_CXVSCROLL)+
		 TOOLS_WIDTH;

  WindowHeight = (rText.bottom-rText.top+1)+
		 GetSystemMetrics(SM_CYFRAME)*2+
		 GetSystemMetrics(SM_CYCAPTION)+
		 GetSystemMetrics(SM_CYMENU)+
		 STATUS_HEIGHT;

  GetWindowRect(hWndMain, &rWind);

  MoveWindow(hWndMain, rWind.top, rWind.left, WindowWidth, WindowHeight, FALSE);

  /* Show the Window */
  ShowWindow(hWndMain, nCmdShow);
  UpdateWindow(hWndMain);

  return TRUE;
}

/* --------------------------------------------------------------------------
 * Main Window WinProc:
 * ------------------------------------------------------------------------*/

LRESULT FAR APIENTRY _export MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
#if BCWIN32
	case WM_CTLCOLORBTN:			/* 3-D look		   */
	case WM_CTLCOLORDLG:
	case WM_CTLCOLOREDIT:
	case WM_CTLCOLORLISTBOX:
	case WM_CTLCOLORMSGBOX:
	case WM_CTLCOLORSCROLLBAR:
	case WM_CTLCOLORSTATIC:
#else
	case WM_SYSCOLORCHANGE:
#endif
	    Ctl3dColorChange();
	    break;

	case WM_LBUTTONDOWN:
	    DoLButtonDown(hWnd, msg, wParam, lParam);
	    break;

	case WM_LBUTTONUP:
	    DoLButtonUp(hWnd, msg, wParam, lParam);
	    break;

	case WM_MOUSEMOVE:
	    DoMouseMove(hWnd, msg, wParam, lParam);
	    break;

	case WM_SIZE:
	    DoSize(hWnd, msg, wParam, lParam);
	    break;

	case WM_GETMINMAXINFO:
	    DoGetMinMaxInfo(hWnd, msg, wParam, lParam);
	    break;

	case WM_CREATE:
	    return ((LONG) DoCreate(hWnd, msg, wParam, lParam));

	case WM_SETFOCUS:
	    SetFocus(hWndText);
	    break;

	case WM_PAINT:
	    DoPaint(hWnd, msg, wParam, lParam);
	    break;

	case WM_COMMAND:
	    DoCommand(hWnd, msg, wParam, lParam);
	    break;

	case WM_MENUSELECT:
	    DoMenuSelect(hWnd, msg, wParam, lParam);
	    break;

	case WM_DESTROY:
	    DoDestroy(hWnd, msg, wParam, lParam);
	    break;

	case WM_CLOSE:
	    everybody(EXIT);
	    DestroyWindow(hWnd);
	    exit(0);
	    break;

	case WM_INITMENU:
	    DoInitMenu(hWnd, msg, wParam, lParam);
	    break;

	default:
	    return DefWindowProc(hWnd, msg, wParam, lParam);
    }
    return (LONG) FALSE;
}

#if BCWIN32
#define DlgSendMessage(h,c,w,l) SendMessage((h),(c),MAKEWPARAM(w,(HIWORD(l))),(LOWORD(l)))
#else
#define DlgSendMessage(h,c,w,l) SendMessage((h),(c),(w),(l))
#endif

/* --------------------------------------------------------------------------
 * About box:
 * ------------------------------------------------------------------------*/

static VOID local DoAbout() {
    DLGPROC lpDlg = (DLGPROC)MakeProcInstance((FARPROC)AboutDlgProc,
					      hThisInstance);
    DialogBox(hThisInstance,MAKEINTRESOURCE(ABOUTDLGBOX),GetFocus(),lpDlg);
    FreeProcInstance(lpDlg);
}

BOOL FAR APIENTRY _export AboutDlgProc(HWND hDlg, UINT msg,
				       WPARAM wP, LPARAM lP) {
    switch (msg) {
	case WM_INITDIALOG:
	    return TRUE;

	case WM_COMMAND:
	    if (CMDitem(wP,lP) == IDOK)
		EndDialog(hDlg,TRUE);
	    return TRUE;
    }
    return FALSE;
}

/* --------------------------------------------------------------------------
 * Script Manager:
 * ------------------------------------------------------------------------*/

static INT smLoaded, smUpto;
static String smFile[NUM_SCRIPTS];
static INT selScr;

static void local SmSelScr(HWND hDlg, Int i) {
    selScr = i;
    SendDlgItemMessage(hDlg, LB_SCRIPTS, LB_SETCURSEL, i, 0L);
}

static void local SmAddScr(HWND hDlg, CHAR *s) {
    smFile[smUpto] = strCopy(s);
    fprintf(stdstr,"%s\n",smFile[smUpto]);
    SendDlgItemMessage(hDlg, LB_SCRIPTS, LB_ADDSTRING, 0,
		       (LONG) (LPSTR) stdstrbuff);
    SmSelScr(hDlg,smUpto);
    smUpto++;
}

#pragma argsused
BOOL CALLBACK ScriptManDlgProc(HWND hDlg, UINT msg,
			       WPARAM wParam, LPARAM lParam) {
    switch (msg) {
	case WM_INITDIALOG: {
	    Int i;
	    smLoaded = numScripts;
	    smUpto   = 0;
	    SendDlgItemMessage(hDlg,LB_SCRIPTS,WM_SETREDRAW,FALSE,0L);
	    for (i=0; i<namesUpto; i++)
		SmAddScr(hDlg,scriptName[i]);
	    SmSelScr(hDlg,-1);
	    SendDlgItemMessage(hDlg,LB_SCRIPTS,WM_SETREDRAW,TRUE,0L);
	    return TRUE;
	}

	case WM_COMMAND:
	    switch (CMDitem(wParam,lParam)) {
		case ID_ADDSCRIPT:
		    if (smUpto>=NUM_SCRIPTS)
			MessageBox(hDlg,"Too many script files",
				   "Add script", MB_ICONEXCLAMATION|MB_OK);
		    else {
			String s = GetaFileName(hDlg,IDS_FILTERFILE);
			if (s)
			    SmAddScr(hDlg,s);
		    }
		    return TRUE;

		case ID_DELSCRIPT:
		    if (selScr < 0)
			MessageBox(hDlg,"No script file selected",
				   "Remove script", MB_ICONEXCLAMATION|MB_OK);
		    else if (selScr == 0)
			MessageBox(hDlg,"Cannot remove prelude file",
				   "Remove script", MB_ICONEXCLAMATION|MB_OK);
		    else {
			Int i;
			SendDlgItemMessage(hDlg, LB_SCRIPTS, LB_DELETESTRING,
					   selScr, 0L);
			if (selScr<smLoaded)
			    smLoaded = selScr;
			if (smFile[selScr]) {
			    free(smFile[selScr]);
			    smFile[selScr] = 0;
			}
			for (i=selScr+1; i<smUpto; ++i)
			    smFile[i-1] = smFile[i];
			smUpto--;
			SmSelScr(hDlg,-1);
		    }
		    return TRUE;

		case ID_EDITSCRIPT:
		    if (selScr >= 0)
			SendMessage(hDlg, WM_COMMAND, LB_SCRIPTS,
				    MAKELONG(0, LBN_DBLCLK));
		    else
			MessageBox(hDlg,"No file selected","Edit",
				   MB_ICONEXCLAMATION|MB_OK);
		    return TRUE;

		case ID_CLEARSCRIPTS: {
		    Int i;
		    for (i=1; i<smUpto; ++i)
			if (smFile[i])
			    free(smFile[i]);
		    smUpto = smLoaded = 1;
		    SendDlgItemMessage(hDlg,LB_SCRIPTS,LB_RESETCONTENT,0,0L);
		    fprintf(stdstr,"%s\n",smFile[0]);
		    SendDlgItemMessage(hDlg, LB_SCRIPTS, LB_ADDSTRING, 0,
				       (LONG) (LPSTR) stdstrbuff);
		    SmSelScr(hDlg,-1);
		    return TRUE;
		}

		case LB_SCRIPTS:
		    switch (CMDdata(wParam,lParam)) {
			case LBN_SELCHANGE:
			    SmSelScr(hDlg,(Int)SendDlgItemMessage(hDlg,
								  LB_SCRIPTS,
								  LB_GETCURSEL,
								  0, 0L));
			    return TRUE;

			case LBN_DBLCLK:
			    SendDlgItemMessage(hDlg,
					       LB_SCRIPTS,
					       LB_GETTEXT,
					       selScr,
					       (LPARAM) (LPSTR) buffer);
			    setLastEdit(buffer,0);
			    runEditor();
			    return TRUE;
		    }
		    break;

		case IDOK: {
		    Int i;
		    for (i=0; i<namesUpto; i++)
			if (scriptName[i])
			    free(scriptName[i]);
		    for (i=0; i<smUpto; i++) {
			scriptName[i] = smFile[i];
			smFile[i]     = 0;
		    }
		    namesUpto  = smUpto;
		    numScripts = smLoaded;
		    dropModulesFrom(numScripts-1);
		    PostMessage(hWndMain,WM_COMMAND,ID_RELOAD,0L);
		    EndDialog(hDlg,TRUE);
		    return TRUE;
		}

		case IDCANCEL: {
		    Int i;
		    for (i=0; i<smUpto; i++)
			if (smFile[i])
			    free(smFile[i]);
		    EndDialog(hDlg, TRUE);
		    return TRUE;
		}
	    }
	    break;
    }
    return FALSE;
}

/* --------------------------------------------------------------------------
 * Options:
 * ------------------------------------------------------------------------*/

static VOID local DoOptions() {
    DLGPROC lpDlg = (DLGPROC)MakeProcInstance((FARPROC)OptionsDlgProc,
					      hThisInstance);
    DialogBox(hThisInstance,MAKEINTRESOURCE(OPTIONSDLGBOX),GetFocus(),lpDlg);
    FreeProcInstance(lpDlg);
}

#pragma argsused
BOOL FAR APIENTRY _export OptionsDlgProc(HWND hDlg, UINT msg,
					 WPARAM wP, LPARAM lP) {
    switch (msg) {
	case WM_INITDIALOG: {
	    Int i = 0;
	    for (; toggle[i].c; i++)
		CheckDlgButton(hDlg,ID_OP+i,*toggle[i].flag);
	    SetDlgItemText(hDlg,ID_PROMPT,  (LPCSTR)prompt);
	    SetDlgItemText(hDlg,ID_LASTEXPR,(LPCSTR)repeatStr);
	    SetDlgItemText(hDlg,ID_HUGSEDIT,(LPCSTR)hugsEdit);
	    SetDlgItemText(hDlg,ID_HUGSPATH,(LPCSTR)hugsPath);
	    return TRUE;
	}

	case WM_COMMAND :
	    switch (CMDitem(wP,lP)) {

		case IDCANCEL :
		    EndDialog(hDlg, TRUE);
		    return FALSE;

		case IDOK : {
		    Int i=0;
		    for (; toggle[i].c; i++)
			*toggle[i].flag
			    = (Bool)IsDlgButtonChecked(hDlg,ID_OP+i);

		    GetDlgItemText(hDlg,ID_PROMPT,(LPSTR)buffer,MAXBUF);
		    if (*buffer && buffer[strlen(buffer)-1]!=' ')
			strcat(buffer," ");
		    if (prompt) free(prompt);
		    prompt = strCopy(buffer);

		    GetDlgItemText(hDlg,ID_LASTEXPR,(LPSTR)buffer,MAXBUF);
		    if (repeatStr) free(repeatStr);
		    repeatStr = strCopy(buffer);

		    GetDlgItemText(hDlg,ID_HUGSPATH,(LPSTR)buffer,MAXBUF);
		    if (hugsPath) free(hugsPath);
		    hugsPath = strCopy(buffer);

		    GetDlgItemText(hDlg,ID_HUGSEDIT,(LPSTR)buffer,MAXBUF);
		    if (hugsEdit) free(hugsEdit);
		    hugsEdit = strCopy(buffer);

		    EndDialog(hDlg, TRUE);
		    return TRUE;
		}

		default	:
		    return TRUE;
	    }
    }
    return FALSE;
}

/* --------------------------------------------------------------------------
 * Loading and saving options to "Hugs.ini":
 * ------------------------------------------------------------------------*/

static CHAR inifile[] = "Hugs.ini";
static CHAR optsec[]  = "Options";

static void local LoadProfileOptions(VOID) {
    INT  NewHeapSize, NewCols, NewRows;

    GetPrivateProfileString("Hugs", "Path", "c:\\hugs\\lib",
			    buffer, MAXBUF, inifile);
    if (hugsPath) free(hugsPath);
    hugsPath = strCopy(buffer);

    GetPrivateProfileString("Hugs", "Edit", "c:\\windows\notepad.exe",
			    buffer, MAXBUF, inifile);
    if (hugsEdit) free(hugsEdit);
    hugsEdit = strCopy(buffer);

    GetPrivateProfileString(optsec, "Prompt", "? ", buffer, MAXBUF, inifile);
    if (*buffer && buffer[strlen(buffer)-1]!=' ')
    	strcat(buffer," ");
    if (prompt) free(prompt);
    prompt = strCopy(buffer);

    GetPrivateProfileString(optsec, "RepeatStr", "$$", buffer, MAXBUF, inifile);
    if (repeatStr) free(repeatStr);
    repeatStr = strCopy(buffer);

    showStats       = GetPrivateProfileInt(optsec, "s", showStats,         inifile);
    addType         = GetPrivateProfileInt(optsec, "t", addType,           inifile);
    failOnError     = GetPrivateProfileInt(optsec, "f", failOnError,       inifile);
    gcMessages      = GetPrivateProfileInt(optsec, "g", gcMessages,        inifile);
    literateScripts = GetPrivateProfileInt(optsec, "l", literateScripts,   inifile);
    literateErrors  = GetPrivateProfileInt(optsec, "e", literateErrors,    inifile);
    useDots         = GetPrivateProfileInt(optsec, ".", useDots,           inifile);
    listFiles       = GetPrivateProfileInt(optsec, "w", listFiles,         inifile);
    kindExpert      = GetPrivateProfileInt(optsec, "k", kindExpert,        inifile);
    useShow         = GetPrivateProfileInt(optsec, "u", useShow,	   inifile);
    chaseImports    = GetPrivateProfileInt(optsec, "i", chaseImports,	   inifile);

    NewHeapSize = GetPrivateProfileInt("Heap", "Size", DEFAULTHEAP, inifile);

    if (heapBuilt() && NewHeapSize != heapSize) {
	MessageBox(GetFocus(),
		   "To change Heap size \n run Hugs again",
		   appName,
		   MB_OK|MB_ICONINFORMATION);
    }
    else {
	heapSize = NewHeapSize;
	if (heapSize<MINIMUMHEAP) {
	    wsprintf(buffer,
		     "Heap size in Hugs.Ini too small. Using %d bytes",
		     DEFAULTHEAP);
	    MessageBox(GetFocus(), buffer, appName, MB_OK|MB_ICONINFORMATION);
	    heapSize = DEFAULTHEAP;
    }
   #ifdef MAXIMUMHEAP
   #if    MAXIMUMHEAP
   else if (heapSize>MAXIMUMHEAP) {
     wsprintf(buffer, "Heap size in Hugs.Ini too small. Using %d bytes",
	      MAXIMUMHEAP);
     MessageBox(GetFocus(), buffer, appName, MB_OK|MB_ICONINFORMATION);
     heapSize = MAXIMUMHEAP;
   }
   #endif
   #endif
  }

  NewRows = GetPrivateProfileInt("Screen", "Rows",    ROWS, inifile);
  NewCols = GetPrivateProfileInt("Screen", "Columns", COLS, inifile);

  if (ScreenRows && NewRows != ScreenRows) {
   MessageBox(GetFocus(),
	      "To change number of screen rows run Hugs again",
	      appName, MB_OK|MB_ICONINFORMATION);
  }
  else
      ScreenRows = NewRows;

  if (ScreenCols && NewCols != ScreenCols) {
   MessageBox(GetFocus(),
	      "To change number of screen columns run Hugs again",
	      appName, MB_OK|MB_ICONINFORMATION);
  }
  else
      ScreenCols = NewCols;
}

static void local SaveProfileOptions(VOID) {
  WritePrivateProfileString("Hugs", "Path", hugsPath, inifile);
  WritePrivateProfileString("Hugs", "Edit", hugsEdit, inifile);
  WritePrivateProfileString(optsec, "Prompt", prompt, inifile);
  WritePrivateProfileString(optsec, "RepeatStr", repeatStr, inifile);

  wsprintf(buffer, "%d", showStats);
  WritePrivateProfileString(optsec, "s", buffer, inifile);
  wsprintf(buffer, "%d",addType);
  WritePrivateProfileString(optsec, "t", buffer, inifile);
  wsprintf(buffer, "%d", failOnError);
  WritePrivateProfileString(optsec, "f", buffer, inifile);
  wsprintf(buffer, "%d", gcMessages);
  WritePrivateProfileString(optsec, "g", buffer, inifile);
  wsprintf(buffer, "%d", literateScripts);
  WritePrivateProfileString(optsec, "l", buffer, inifile);
  wsprintf(buffer, "%d", literateErrors);
  WritePrivateProfileString(optsec, "e", buffer, inifile);
  wsprintf(buffer, "%d", useDots);
  WritePrivateProfileString(optsec, ".", buffer, inifile);
  wsprintf(buffer, "%d", listFiles);
  WritePrivateProfileString(optsec, "w", buffer, inifile);
  wsprintf(buffer, "%d", kindExpert);
  WritePrivateProfileString(optsec, "k", buffer, inifile);
  wsprintf(buffer, "%d", useShow);
  WritePrivateProfileString(optsec, "u", buffer, inifile);
  wsprintf(buffer, "%d", chaseImports);
  WritePrivateProfileString(optsec, "i", buffer, inifile);

  wsprintf(buffer, "%d", heapSize);
  WritePrivateProfileString("Heap", "Size", buffer, inifile);

  wsprintf(buffer, "%d", ScreenRows);
  WritePrivateProfileString("Screen", "Rows", buffer, inifile);
  wsprintf(buffer, "%d", ScreenCols);
  WritePrivateProfileString("Screen", "Columns", buffer, inifile);

  /* Flush to disk now*/
  WritePrivateProfileString(NULL, NULL, NULL, inifile);
}

/* --------------------------------------------------------------------------
 * Browser dialog boxes:
 * ------------------------------------------------------------------------*/

static local 	void SetClass		Args((HWND, Class));
static local 	void SetName		Args((HWND, UINT, List));
static local 	void SetTycon		Args((HWND, UINT, List));

/* When a class changes to currClass get new list of instances, */
/* members and contexrs for the new class                       */

static local VOID SetClass(HWND hDlg, Class currClass) {
    List instances = cclass(currClass).instances;
    List members   = cclass(currClass).members;
    List supers    = cclass(currClass).supers;

    /* Update list of instances */
    SendDlgItemMessage(hDlg,LB_INSTANCES,LB_RESETCONTENT,0,0L);
    for (; nonNull(instances); instances=tl(instances)) {
	SendDlgItemMessage(hDlg,LB_INSTANCES,LB_ADDSTRING,0,
			   (LONG)hd(instances));
	SendDlgItemMessage(hDlg,LB_INSTANCES,LB_SETCURSEL,0,0L);
    }

    /* Update list of members */
    SendDlgItemMessage(hDlg,LB_MEMBERS,LB_RESETCONTENT,0,0L);
    for (; nonNull(members); members=tl(members)) {
	SendDlgItemMessage(hDlg,LB_MEMBERS,LB_ADDSTRING,0,
			   (LONG)hd(members));
	SendDlgItemMessage(hDlg,LB_MEMBERS,LB_SETCURSEL,0,0L);
    }

    /* Update superclasses */
    SendDlgItemMessage(hDlg,LB_CONTEXT,LB_RESETCONTENT,0,0L);
    for (; nonNull(supers); supers=tl(supers)) {
	fprintf(stdstr,"%s",textToStr(cclass(hd(supers)).text));
	if (nonNull(tl(supers)))
	    fprintf(stdstr,", ");
    }
    fprintf(stdstr,"\n");
    SendDlgItemMessage(hDlg,LB_CONTEXT,LB_ADDSTRING,0,(LONG)(LPSTR)stdstrbuff);
}

/* Handles browse classes dialog box */
#pragma argsused
BOOL FAR APIENTRY _export BrowseClassesDlgProc(HWND hDlg, UINT msg,
					       WPARAM wParam, LPARAM lParam) {
    static Class currClass;
    Class	 theClass;
    Inst	 theInst;
    Name	 theMember;
    Module       module;
    HBITMAP      hBitmap;
    RECT         aRect, DlgRect;
    HBITMAP      hBmp, hsvBmp1, hsvBmp2, hsvBmp3, hsvBmp4, hsvBmp5, hsvBmp6;
    BITMAP	 bm;
    static HDC   hDCMemory, hCDC, hCDCSel, hIDC, hIDCSel, hMDC, hMDCSel;

    switch (msg) {

	case WM_INITDIALOG: {/* Create list of classes and set current class */
	    Int i;
	    for (i=CLASSMIN; i<classMax(); i++)
		SendDlgItemMessage(hDlg, LB_CLASS ,LB_ADDSTRING,
						0, (LPARAM)(LPSTR) i);
	    SendDlgItemMessage(hDlg, LB_CLASS, LB_SETCURSEL, 0, 0L);
	    currClass = (Class)SendDlgItemMessage(hDlg,
					       LB_CLASS,
					       LB_GETITEMDATA,
					       (WPARAM)SendDlgItemMessage(hDlg, LB_CLASS ,LB_GETCURSEL, 0, 0L), 0L);
	    SetClass(hDlg, currClass);

	    /* Create Bitmaps */
	    hsvBmp1 = CreatehDC(hDlg, &hCDC,    "CLASSBMP",    GetSysColor(COLOR_WINDOW));
	    hsvBmp2 = CreatehDC(hDlg, &hCDCSel, "CLASSBMP",    GetSysColor(COLOR_HIGHLIGHT));
	    hsvBmp3 = CreatehDC(hDlg, &hIDC,    "INSTANCEBMP", GetSysColor(COLOR_WINDOW));
	    hsvBmp4 = CreatehDC(hDlg, &hIDCSel, "INSTANCEBMP", GetSysColor(COLOR_HIGHLIGHT));
	    hsvBmp5 = CreatehDC(hDlg, &hMDC,    "MEMBERBMP",   GetSysColor(COLOR_WINDOW));
	    hsvBmp6 = CreatehDC(hDlg, &hMDCSel, "MEMBERBMP",   GetSysColor(COLOR_HIGHLIGHT));
	    break;
	}

	case WM_DESTROY: /* Destroy Bitmaps */
	    SelectObject(hCDC, hsvBmp1);     DeleteDC(hCDC);
	    SelectObject(hCDCSel, hsvBmp2);  DeleteDC(hCDCSel);
	    SelectObject(hIDC, hsvBmp3);     DeleteDC(hIDC);
	    SelectObject(hIDCSel, hsvBmp4);  DeleteDC(hIDCSel);
	    SelectObject(hMDC, hsvBmp5);     DeleteDC(hMDC);
	    SelectObject(hMDCSel, hsvBmp6);  DeleteDC(hMDCSel);
	    break;

	case WM_PAINT: /* Paint classes Bitmap */
	    GetWindowRect(hDlg, &DlgRect);
	    GetWindowRect(GetDlgItem(hDlg,ID_BROWSEBMP),&aRect);

	    hBitmap = LoadBitmap(hThisInstance, "CLASSESDLGBMP");
	    PutBitmap(hBitmap,
		      hDlg,
		      aRect.left-DlgRect.left-GetSystemMetrics(SM_CXDLGFRAME),
		      aRect.top-DlgRect.top-GetSystemMetrics(SM_CYDLGFRAME)
				-GetSystemMetrics(SM_CYCAPTION),
		      SRCCOPY);
	    DeleteObject(hBitmap);
	    break;

	case WM_COMPAREITEM: {
	    LPCOMPAREITEMSTRUCT lpcis = (LPCOMPAREITEMSTRUCT) lParam;
	    switch (wParam) {
		case LB_CLASS     :
		case LB_INSTANCES :
		case LB_MEMBERS   : return (lpcis->itemData1==lpcis->itemData2);
	    }
	    break;
	}

	case WM_MEASUREITEM: {
	    DRAWITEMSTRUCT FAR *lpdis = (DRAWITEMSTRUCT FAR *) lParam;
	    if (lpdis->CtlID == LB_CLASS || lpdis->CtlID == LB_INSTANCES ||
		lpdis->CtlID == LB_MEMBERS ) {
		LPMEASUREITEMSTRUCT lpmis = (LPMEASUREITEMSTRUCT) lParam;

		/* Set the height of the list box items to Bitmap height */
		hBmp = LoadBitmap(hThisInstance, "CLASSBMP");
		GetObject(hBmp, sizeof(BITMAP), &bm);
		DeleteObject(hBmp);
		lpmis->itemHeight = bm.bmHeight+1;
		return TRUE;
	    }
	    break;
	}

	case WM_DRAWITEM: {
	    DRAWITEMSTRUCT FAR *lpdis = (DRAWITEMSTRUCT FAR *) lParam;
	    if (lpdis->CtlID == LB_CLASS || lpdis->CtlID == LB_INSTANCES ||
		lpdis->CtlID == LB_MEMBERS ) {
		Bool selected = FALSE;

		if (lpdis->itemID == (UINT)(-1))
		    return TRUE;

		switch (lpdis->itemAction) {
		    case ODA_DRAWENTIRE:
		    case ODA_SELECT:
		    case ODA_FOCUS:
			if ((lpdis->itemState & ODS_SELECTED)) {
			    SetBkColor(lpdis->hDC, GetSysColor(COLOR_HIGHLIGHT));
			    SetTextColor(lpdis->hDC, GetSysColor(COLOR_HIGHLIGHTTEXT));
			    selected = TRUE;
			}
			else {
			    SetBkColor(lpdis->hDC, GetSysColor(COLOR_WINDOW));
			    SetTextColor(lpdis->hDC, GetSysColor(COLOR_WINDOWTEXT));
			}
			break;

		    default:
			return FALSE;
		}

		/* Get bitmap size */
		hBmp = LoadBitmap(hThisInstance, "CLASSBMP");
		GetObject(hBmp, sizeof(BITMAP), &bm);
		DeleteObject(hBmp);

		switch (lpdis->CtlID) {
		    case LB_CLASS:
			theClass = (Class) SendDlgItemMessage(hDlg, lpdis->CtlID, LB_GETITEMDATA, lpdis->itemID, 0);
			fprintf(stdstr,"%s\n",textToStr(cclass(theClass).text));
			ExtTextOut(lpdis->hDC, lpdis->rcItem.left+21, lpdis->rcItem.top, ETO_OPAQUE, &(lpdis->rcItem), stdstrbuff, strlen(stdstrbuff), NULL);
			hDCMemory = selected ? hCDCSel : hCDC;
			break;

		    case LB_INSTANCES:
			theInst = (Inst) SendDlgItemMessage(hDlg, lpdis->CtlID, LB_GETITEMDATA, lpdis->itemID, 0);
			printPred(stdstr,makeInstPred(theInst));
			fprintf(stdstr,"\n");
			ExtTextOut(lpdis->hDC, lpdis->rcItem.left+21, lpdis->rcItem.top, ETO_OPAQUE, &(lpdis->rcItem), stdstrbuff, strlen(stdstrbuff), NULL);
			hDCMemory = selected ? hIDCSel : hIDC;
			break;

		    case LB_MEMBERS:
			theMember = (Name) SendDlgItemMessage(hDlg, lpdis->CtlID, LB_GETITEMDATA, lpdis->itemID, 0);
			printExp(stdstr, theMember);
			fprintf(stdstr, " :: ");
			printType(stdstr, name(theMember).type);
			fprintf(stdstr,"\n");
			ExtTextOut(lpdis->hDC, lpdis->rcItem.left+21, lpdis->rcItem.top, ETO_OPAQUE, &(lpdis->rcItem), stdstrbuff, strlen(stdstrbuff), NULL);
			hDCMemory = selected ? hMDCSel : hMDC;
			break;
		}

		BitBlt(lpdis->hDC, (lpdis->rcItem.left)+4, (lpdis->rcItem.top),
				   bm.bmWidth,             bm.bmHeight,
				   hDCMemory,
				   0,
				   0,
				   SRCCOPY);

		/* If selected draw rectangle */
		if ((lpdis->itemState & ODS_SELECTED)&&(lpdis->itemState & ODS_FOCUS))
		    DrawFocusRect(lpdis->hDC, &(lpdis->rcItem));

		return TRUE;
	    }
	    break;
	}

	case WM_COMMAND:
	    switch (CMDitem(wParam,lParam)) {
		case LB_CLASS:
		    switch(CMDdata(wParam,lParam)) {
			case LBN_SELCHANGE: /* select a new class */
			    currClass = (Class) SendDlgItemMessage(hDlg, LB_CLASS ,LB_GETITEMDATA, SendDlgItemMessage(hDlg, LB_CLASS ,LB_GETCURSEL, 0, 0L), 0L);
			    SetClass(hDlg, currClass);
			    break;

			case LBN_DBLCLK: /* Open in text editor script file with class definition */
			    currClass = (Class) SendDlgItemMessage(hDlg, LB_CLASS ,LB_GETITEMDATA, SendDlgItemMessage(hDlg, LB_CLASS ,LB_GETCURSEL, 0, 0L), 0L);
			    module = startNewModule();
			    setLastEdit(scriptName[moduleThisClass(currClass)], cclass(currClass).line);
			    dropModulesFrom(module);
			    runEditor();
			    break;
		    }
		    break;

		case LB_MEMBERS:
		case LB_INSTANCES:
		    switch(CMDdata(wParam,lParam)) {/* Open in text editor script file with instance definition */
			case LBN_DBLCLK: {
			    Inst  currInst;
			    currInst = (Inst) SendDlgItemMessage(hDlg, LB_INSTANCES, LB_GETITEMDATA, SendDlgItemMessage(hDlg, LB_INSTANCES ,LB_GETCURSEL, 0, 0L), 0L);

			    /* Find instance module */
			    module = startNewModule();
			    setLastEdit(scriptName[moduleThisInst(currInst)], inst(currInst).line);
			    dropModulesFrom(module);
			    runEditor();
			    break;
			}
		    }
		    break;

		case ID_HIERARCHY: /* Draw class hierarchy */
		    Hierarchy();
		    break;

		case ID_EDITCLASS: /* Pushed on Edit class button */
		    if (SendDlgItemMessage(hDlg, LB_CLASS, LB_GETCURSEL, 0, 0L) != LB_ERR)
			DlgSendMessage(hDlg, WM_COMMAND, LB_CLASS, MAKELONG(0, LBN_DBLCLK));
		    break;

		case ID_EDITINSTANCE: /* Pushed on Edit instance button */
		    if (SendDlgItemMessage(hDlg, LB_INSTANCES, LB_GETCURSEL, 0, 0L) != LB_ERR)
			DlgSendMessage(hDlg, WM_COMMAND, LB_INSTANCES, MAKELONG(0, LBN_DBLCLK));
		    break;

		case IDCANCEL: /* Close dialog */
		case IDOK:
		    EndDialog(hDlg, TRUE);
		    return TRUE;

		default:
		    return TRUE;
	    }
    }
    return FALSE;
}

/* When the name  selected changes to currName get its type and definition */
static local VOID SetName(HWND hDlg, UINT currName, List names) {
    List nameList = names;
    Name nm;
    UINT i;

    for (i=0; i<currName; nameList=tl(nameList))
	i++;

    nm = hd(nameList);

    if (isCfun(nm))
	fprintf(stdstr,"Data constructor, ");
    else if (isMfun(nm))
	fprintf(stdstr, "Class member, ");
    else if (isSfun(nm))
	fprintf(stdstr, "Selector function, ");
    else if (name(nm).primDef)
	fprintf(stdstr, "Primitive, ");

    printExp(stdstr,nm);
    fprintf(stdstr," :: ");
    if (nonNull(name(nm).type))
	printType(stdstr,name(nm).type);
    else
      fprintf(stdstr,"<Unknown type>");
    if (name(nm).line)
	fprintf(stdstr,"   (%s,%d)",scriptName[moduleThisName(nm)],name(nm).line);
    fprintf(stdstr, "\n");

    SendDlgItemMessage(hDlg,LB_NAMESNOTES,LB_RESETCONTENT,0,0L);
    SendDlgItemMessage(hDlg,LB_NAMESNOTES,LB_ADDSTRING,0,
		       (LONG)(LPSTR)stdstrbuff);
}

#pragma argsused
/* Handles browse names dialog box */
BOOL FAR APIENTRY _export BrowseNamesDlgProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
 static              List nameList = NIL;
 List                names=NIL;
 Name		     nm;
 UINT                theName;
 Module              module;
 WORD          	     NotifyCode = CMDdata(wParam,lParam);
 WORD		     wId        = CMDitem(wParam,lParam);
 RECT                aRect, DlgRect;
 HBITMAP             hBitmap;
 HBITMAP             hBmp, hsvBmp0, hsvBmp1, hsvBmp2, hsvBmp3, hsvBmp4,
		     hsvBmp5, hsvBmp6, hsvBmp7, hsvBmp8, hsvBmp9;
 BITMAP	             bm;
 static HDC          hDCMemory, hPDC, hPDCSel, hDDC, hDDCSel, hMDC, hMDCSel,
		     hNDC, hNDCSel, hSDC, hSDCSel;
 DRAWITEMSTRUCT FAR *lpdis;
 LPMEASUREITEMSTRUCT lpmis;
 BOOL                Selected = FALSE;

 switch (msg) {

     case WM_INITDIALOG:
	  SendDlgItemMessage(hDlg, LB_NAMES, LB_SETCOLUMNWIDTH, (WPARAM)(100), 0L);
	  nameList = addNamesMatching((String)0, NIL);

	  for (names=nameList; nonNull(names); names=tl(names)) {
	   if (nonNull(names)) {
	    fprintf(stdstr, "%s\n",textToStr(name(hd(names)).text));
	    SendDlgItemMessage(hDlg, LB_NAMES ,LB_ADDSTRING, 0, (LONG)(LPSTR) stdstrbuff);
	    SendDlgItemMessage(hDlg, LB_NAMES, LB_SETCURSEL, 0, 0L);
	   }
	  }
	  theName = 0;
	  SetName(hDlg, theName, nameList);

	  hsvBmp0 = CreatehDC(hDlg, &hPDC,     "PRIMBMP", GetSysColor(COLOR_WINDOW));
	  hsvBmp1 = CreatehDC(hDlg, &hPDCSel,  "PRIMBMP", GetSysColor(COLOR_HIGHLIGHT));
	  hsvBmp2 = CreatehDC(hDlg, &hDDC,     "DATACONSBMP", GetSysColor(COLOR_WINDOW));
	  hsvBmp3 = CreatehDC(hDlg, &hDDCSel,  "DATACONSBMP", GetSysColor(COLOR_HIGHLIGHT));
	  hsvBmp4 = CreatehDC(hDlg, &hMDC,     "MEMBERBMP", GetSysColor(COLOR_WINDOW));
	  hsvBmp5 = CreatehDC(hDlg, &hMDCSel,  "MEMBERBMP", GetSysColor(COLOR_HIGHLIGHT));
	  hsvBmp6 = CreatehDC(hDlg, &hNDC,     "NAMEBMP", GetSysColor(COLOR_WINDOW));
	  hsvBmp7 = CreatehDC(hDlg, &hNDCSel,  "NAMEBMP", GetSysColor(COLOR_HIGHLIGHT));
	  hsvBmp8 = CreatehDC(hDlg, &hSDC,     "SELECTORBMP", GetSysColor(COLOR_WINDOW));
	  hsvBmp9 = CreatehDC(hDlg, &hSDCSel,  "SELECTORBMP", GetSysColor(COLOR_HIGHLIGHT));
	  break;

     case WM_DESTROY:
	  SelectObject(hPDC,    hsvBmp0); DeleteDC(hPDC);
	  SelectObject(hPDCSel, hsvBmp1); DeleteDC(hPDCSel);
	  SelectObject(hDDC,    hsvBmp2); DeleteDC(hDDC);
	  SelectObject(hDDCSel, hsvBmp3); DeleteDC(hDDCSel);
	  SelectObject(hMDC,    hsvBmp4); DeleteDC(hMDC);
	  SelectObject(hMDCSel, hsvBmp5); DeleteDC(hMDCSel);
	  SelectObject(hNDC,    hsvBmp6); DeleteDC(hNDC);
	  SelectObject(hNDCSel, hsvBmp7); DeleteDC(hNDCSel);
	  SelectObject(hSDC,    hsvBmp8); DeleteDC(hSDC);
	  SelectObject(hSDCSel, hsvBmp9); DeleteDC(hSDCSel);
	  break;

     case WM_MEASUREITEM:
	 lpdis = (DRAWITEMSTRUCT FAR *) lParam;

	 if (lpdis->CtlID == LB_NAMES) {
	     lpmis = (LPMEASUREITEMSTRUCT) lParam;
	     /* Set the height of the list box items to Bitmap height */
	     hBmp = LoadBitmap(hThisInstance, "PRIMBMP");
	     GetObject(hBmp, sizeof(BITMAP), &bm);
	     DeleteObject(hBmp);
	     lpmis->itemHeight = bm.bmHeight+1;
	     return TRUE;
	 }
	 break;

     case WM_DRAWITEM:
	 lpdis = (DRAWITEMSTRUCT FAR *) lParam;
	 if(lpdis->CtlID == LB_NAMES) {
	  if (lpdis->itemID == (UINT)-1) {
	   return TRUE;
	  }

	  switch (lpdis->itemAction) {
	    case ODA_DRAWENTIRE:
	    case ODA_SELECT:
	    case ODA_FOCUS:
		if ((lpdis->itemState & ODS_SELECTED) /*&& (lpdis->itemState & ODS_FOCUS)*/) {
		  SetBkColor(lpdis->hDC, GetSysColor(COLOR_HIGHLIGHT));
		  SetTextColor(lpdis->hDC, GetSysColor(COLOR_HIGHLIGHTTEXT));
		  Selected = TRUE;
		}
		else {
		  SetBkColor(lpdis->hDC, GetSysColor(COLOR_WINDOW));
		  SetTextColor(lpdis->hDC, GetSysColor(COLOR_WINDOWTEXT));
		}
		break;

	    default:
		return FALSE;
	  }

	   SendDlgItemMessage(hDlg, lpdis->CtlID, LB_GETTEXT, lpdis->itemID, (LPARAM)buffer);
	   ExtTextOut(lpdis->hDC, lpdis->rcItem.left+21, lpdis->rcItem.top, (ETO_CLIPPED|ETO_OPAQUE), &(lpdis->rcItem), buffer, strlen(buffer), NULL);

	   /* Get bitmap size */
	   hBmp = LoadBitmap(hThisInstance, "PRIMBMP");
	   GetObject(hBmp, sizeof(BITMAP), &bm);
	   DeleteObject(hBmp);

	   /* Look for name in names table */
	   nm = findName(findText(buffer));
	   if (isCfun(nm))
	       hDCMemory = Selected ? hDDCSel : hDDC;
	   else if (isMfun(nm))
	       hDCMemory = Selected ? hMDCSel : hMDC;
	   else if (isSfun(nm))
	       hDCMemory = Selected ? hSDCSel : hSDC;
	   else if (name(nm).primDef)
	       hDCMemory = Selected ? hPDCSel : hPDC;
	   else
	       hDCMemory = Selected ? hNDCSel : hNDC;

	   BitBlt(lpdis->hDC, (lpdis->rcItem.left)+4, (lpdis->rcItem.top), bm.bmWidth, bm.bmHeight, hDCMemory, 0, 0, SRCCOPY);

	   /* If selected draw rectangle */
	   if ((lpdis->itemState & ODS_SELECTED)&&(lpdis->itemState & ODS_FOCUS)) {
	     DrawFocusRect(lpdis->hDC, &(lpdis->rcItem));
	   }

	   return TRUE;
	 }

  case WM_PAINT:
     GetWindowRect(hDlg, &DlgRect);
     GetWindowRect(GetDlgItem(hDlg, ID_BROWSEBMP), &aRect);

     hBitmap = LoadBitmap(hThisInstance, "NAMESDLGBMP");
     PutBitmap(hBitmap, hDlg, aRect.left-DlgRect.left-GetSystemMetrics(SM_CXDLGFRAME),
	       aRect.top-DlgRect.top-GetSystemMetrics(SM_CYDLGFRAME)-GetSystemMetrics(SM_CYCAPTION),
	       SRCCOPY);
     DeleteObject(hBitmap);
     break;

  case WM_COMMAND:
    switch (wId) {
      case LB_NAMES:
	switch(NotifyCode) {
	  case LBN_SELCHANGE:
	     /* Select a new name */
	     theName = (UINT) SendDlgItemMessage(hDlg, LB_NAMES ,LB_GETCURSEL, 0, 0L);
	     SetName(hDlg, theName, nameList);
	     break;

	  case LBN_DBLCLK: {
	     /* Open in text editor script file with name definition */
	     /* Get the selected name */
	     theName = (UINT) SendDlgItemMessage(hDlg, LB_NAMES ,LB_GETCURSEL, 0, 0L);
	     SendDlgItemMessage(hDlg, LB_NAMES, LB_GETTEXT, theName, (LPARAM) ((LPSTR) buffer));

	     /* Look for it in names table */
	     nm = findName(findText(buffer));

	     if (!name(nm).primDef) {
	       module = startNewModule();
	       setLastEdit(scriptName[moduleThisName(nm)], name(nm).line);
	       dropModulesFrom(module);
	       runEditor();
	     }
	     else {
	       MessageBox(hDlg, "Primitive function:\nNo definition available.", appName, MB_ICONINFORMATION | MB_OK);
	     }
	  }
	  break;
	}
	break;

      case IDC_SEARCHNAME:  /* Search a name */
	switch(HIBYTE(NotifyCode)) {
	  case HIBYTE(EN_CHANGE): {
	    /* Get edit control contents */
	    SendDlgItemMessage(hDlg, IDC_SEARCHNAME, WM_GETTEXT, MAXBUF, (LPARAM) ((LPSTR) buffer));

	    /* Search in names list box */
	    SendDlgItemMessage(hDlg, LB_NAMES, LB_SELECTSTRING, 0, (LPARAM) ((LPSTR) buffer));

	    /* Update window contents */
	    DlgSendMessage(hDlg, WM_COMMAND, LB_NAMES, MAKELONG(0, LBN_SELCHANGE));
	  }
	  break;
	}
	break;

      case ID_EDITNAME: /* Pushed on Edit name button */
	if (SendDlgItemMessage(hDlg, LB_NAMES, LB_GETCURSEL, 0, 0L) != LB_ERR)
	  DlgSendMessage(hDlg, WM_COMMAND, LB_NAMES, MAKELONG(0, LBN_DBLCLK));
	break;

      case IDCANCEL: /* Close dialog */
      case IDOK:
	EndDialog(hDlg, TRUE);
	return TRUE;

      default:
	return TRUE;
    }
 }
 return FALSE;
}

static Int numCfuns;
static Int numSfuns;

/* A new Tycon was selected */
static local VOID SetTycon(HWND hDlg, UINT currTycon, List tycons) {
    List  tyconList = tycons;
    Tycon tc;
    Inst  in;
    UINT  i, j;
    Type  t;

    for (i=0; i<currTycon; tyconList=tl(tyconList))
	i++;
    tc = hd(tyconList);
    numCfuns = 0;
    numSfuns = 0;

    t = tc;
    for (j=0; j<tycon(tc).arity; ++j)
	t = ap(t,mkOffset(j));

    switch(tycon(tc).what) {
	case RESTRICTSYN :
	case SYNONYM     : fprintf(stdstr,"Type synonym");
			   break;
	case NEWTYPE	 : fprintf(stdstr,"Newtype constructor");
			   break;
	case DATATYPE    : fprintf(stdstr,"Datatype constructor");
			   break;
	default          : fprintf(stdstr,"Type constructor");
			   break;
    }

    if (kindExpert) {
	fprintf(stdstr," with kind ");
	printKind(stdstr,tycon(tc).kind);
    }

    fprintf(stdstr, ":\n");
    SendDlgItemMessage(hDlg, LB_CONS ,LB_RESETCONTENT, 0, 0L);
    SendDlgItemMessage(hDlg, LB_CONS ,LB_ADDSTRING, 0, (LONG)(LPSTR) stdstrbuff);

    switch (tycon(tc).what) {
     case SYNONYM:
     fprintf(stdstr, "type ");
     printType(stdstr, t);
     fprintf(stdstr, " = ");
     printType(stdstr, tycon(tc).defn);
     fprintf(stdstr, "\n");
     SendDlgItemMessage(hDlg, LB_CONS ,LB_ADDSTRING, 0, (LONG)(LPSTR) stdstrbuff);
     break;

   case NEWTYPE :
   case DATATYPE: {
     List cs = tycon(tc).defn;
     fprintf  (stdstr,(tycon(tc).what==DATATYPE ? "data " : "newtype "));
     printType(stdstr, t);
     fprintf  (stdstr, " = ...\n");
     SendDlgItemMessage(hDlg, LB_CONS ,LB_ADDSTRING, 0, (LONG)(LPSTR) stdstrbuff);
     for (; nonNull(cs); cs=tl(cs)) {
	 printExp (stdstr, hd(cs));
	 fprintf  (stdstr, " :: ");
	 printType(stdstr, name(hd(cs)).type);
	 fprintf  (stdstr, "\n");
	 SendDlgItemMessage(hDlg, LB_CONS ,LB_ADDSTRING, 0, (LONG)(LPSTR) stdstrbuff);
	 SendDlgItemMessage(hDlg, LB_CONS, LB_SETCURSEL, (WPARAM)(-1), 0L);
	 if (isCfun(hd(cs)))
	     numCfuns++;
	 else
	     numSfuns++;
     }
   }
   break;

   case RESTRICTSYN:
     fprintf  (stdstr, "type");
     printType(stdstr, t);
     fprintf  (stdstr, " = <restricted>\n");
     SendDlgItemMessage(hDlg, LB_CONS ,LB_ADDSTRING, 0, (LONG)(LPSTR) stdstrbuff);
     break;
  }

  for (in=findFirstInst(tc); nonNull(in); in=findNextInst(tc,in)) {
      fprintf(stdstr,"instance ");
      if (nonNull(inst(in).specifics)) {
	  printContext(stdstr,inst(in).specifics);
	  fprintf(stdstr," => ");
      }
      printPred(stdstr,makeInstPred(in));
      fprintf(stdstr,"\n");
      SendDlgItemMessage(hDlg, LB_CONS ,LB_ADDSTRING, 0, (LONG)(LPSTR) stdstrbuff);
      SendDlgItemMessage(hDlg, LB_CONS, LB_SETCURSEL, (WPARAM)(-1), 0L);
  }
  SendDlgItemMessage(hDlg, LB_CONS, LB_SETCURSEL, (WPARAM)(-1), 0L);
}


/* Handles browse Tycons dialog box */
#pragma argsused
BOOL FAR APIENTRY _export BrowseTyconsDlgProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
 static        List tyconList = NIL;
 List          tycons=NIL;
 Tycon         tc;
 UINT          theTycon;
 Module        module;
 WORD          NotifyCode = CMDdata(wParam,lParam);
 WORD	       wId        = CMDitem(wParam,lParam);
 RECT          aRect, DlgRect;
 HBITMAP       hBitmap;
 HBITMAP       hBmp, hsvBmp0, hsvBmp1, hsvBmp2, hsvBmp3, hsvBmp4, hsvBmp5,
	       hsvBmp6, hsvBmp7, hsvBmp8, hsvBmp9;
 BITMAP	       bm;
 static HDC    hDCMemory, hTCDC, hTCDCSel, hDDC, hDDCSel, hTSDC, hTSDCSel,
	       hTNDC, hTNDCSel, hSDC, hSDCSel;
 DRAWITEMSTRUCT FAR *lpdis;
 LPMEASUREITEMSTRUCT lpmis;
 BOOL                Selected = FALSE;
 UINT	       i;

 switch (msg) {

     case WM_INITDIALOG:
	  tyconList = addTyconsMatching((String)0, NIL);
	  for (tycons=tyconList; nonNull(tycons); tycons=tl(tycons)) {
	     tc = hd(tycons);
	     fprintf(stdstr, "%s\n",textToStr(tycon(tc).text));
	     SendDlgItemMessage(hDlg, LB_TYCONS, LB_SETCOLUMNWIDTH, (WPARAM)(100), 0L);
	     SendDlgItemMessage(hDlg, LB_TYCONS ,LB_ADDSTRING, 0, (LONG)(LPSTR) stdstrbuff);
	     SendDlgItemMessage(hDlg, LB_TYCONS, LB_SETCURSEL, 0, 0L);
	  }

	  theTycon = 0;
	  SetTycon(hDlg, theTycon, tyconList);

	  hsvBmp0 = CreatehDC(hDlg, &hTCDC,    "TYPECONSBMP", GetSysColor(COLOR_WINDOW));
	  hsvBmp1 = CreatehDC(hDlg, &hTCDCSel, "TYPECONSBMP", GetSysColor(COLOR_HIGHLIGHT));
	  hsvBmp2 = CreatehDC(hDlg, &hDDC,     "DATACONSBMP", GetSysColor(COLOR_WINDOW));
	  hsvBmp3 = CreatehDC(hDlg, &hDDCSel,  "DATACONSBMP", GetSysColor(COLOR_HIGHLIGHT));
	  hsvBmp4 = CreatehDC(hDlg, &hTSDC,    "TYPESINBMP", GetSysColor(COLOR_WINDOW));
	  hsvBmp5 = CreatehDC(hDlg, &hTSDCSel, "TYPESINBMP", GetSysColor(COLOR_HIGHLIGHT));
	  hsvBmp6 = CreatehDC(hDlg, &hTNDC,    "NEWTYPEBMP", GetSysColor(COLOR_WINDOW));
	  hsvBmp7 = CreatehDC(hDlg, &hTNDCSel, "NEWTYPEBMP", GetSysColor(COLOR_HIGHLIGHT));
	  hsvBmp8 = CreatehDC(hDlg, &hSDC,     "SELECTORBMP", GetSysColor(COLOR_WINDOW));
	  hsvBmp9 = CreatehDC(hDlg, &hSDCSel,  "SELECTORBMP", GetSysColor(COLOR_HIGHLIGHT));
	  break;

     case WM_DESTROY:
	  SelectObject(hTCDC,    hsvBmp0); DeleteDC(hTCDC);
	  SelectObject(hTCDCSel, hsvBmp1); DeleteDC(hTCDCSel);
	  SelectObject(hDDC,     hsvBmp2); DeleteDC(hDDC);
	  SelectObject(hDDCSel,  hsvBmp3); DeleteDC(hDDCSel);
	  SelectObject(hTSDC,    hsvBmp4); DeleteDC(hTSDC);
	  SelectObject(hTSDCSel, hsvBmp5); DeleteDC(hTSDCSel);
	  SelectObject(hTNDC,    hsvBmp6); DeleteDC(hTNDC);
	  SelectObject(hTNDCSel, hsvBmp7); DeleteDC(hTNDCSel);
	  SelectObject(hSDC,     hsvBmp8); DeleteDC(hSDC);
	  SelectObject(hSDCSel,  hsvBmp9); DeleteDC(hSDCSel);
	  break;

     case WM_MEASUREITEM:
	 lpdis = (DRAWITEMSTRUCT FAR *) lParam;

	 if (lpdis->CtlID == LB_TYCONS || lpdis->CtlID == LB_CONS ) {
	  lpmis = (LPMEASUREITEMSTRUCT) lParam;

	  /* Set the height of the list box items to Bitmap height */
	  hBmp = LoadBitmap(hThisInstance, "CLASSBMP");
	  GetObject(hBmp, sizeof(BITMAP), &bm);
	  DeleteObject(hBmp);

	  lpmis->itemHeight = bm.bmHeight+1;
	  return TRUE;
	 }
	 break;

     case WM_DRAWITEM: {
	 Int offs = 0;

	 lpdis = (DRAWITEMSTRUCT FAR *) lParam;

	 if (lpdis->CtlID == LB_TYCONS || lpdis->CtlID == LB_CONS ) {

	  if (lpdis->itemID == (UINT)-1)
	   return TRUE;

	  switch (lpdis->itemAction) {
	    case ODA_DRAWENTIRE:
	    case ODA_SELECT:
	    case ODA_FOCUS:
		if ((lpdis->itemState & ODS_SELECTED) /*&& (lpdis->itemState & ODS_FOCUS)*/) {
		  SetBkColor(lpdis->hDC, GetSysColor(COLOR_HIGHLIGHT));
		  SetTextColor(lpdis->hDC, GetSysColor(COLOR_HIGHLIGHTTEXT));
		  Selected = TRUE;
		}
		else {
		  SetBkColor(lpdis->hDC, GetSysColor(COLOR_WINDOW));
		  SetTextColor(lpdis->hDC, GetSysColor(COLOR_WINDOWTEXT));
		}
		break;

	    default:
		return FALSE;
	  }

	   SendDlgItemMessage(hDlg, lpdis->CtlID, LB_GETTEXT, lpdis->itemID, (LPARAM) buffer);

	   if (lpdis->CtlID == LB_CONS)
	       if (lpdis->itemID <= 1 || lpdis->itemID>=2+numCfuns+numSfuns) {
		   ExtTextOut(lpdis->hDC, lpdis->rcItem.left+4, lpdis->rcItem.top, (ETO_CLIPPED|ETO_OPAQUE), &(lpdis->rcItem), buffer, strlen(buffer), NULL);
		   return TRUE;
	       }
	       else
		   offs = 15;
	   }

	   ExtTextOut(lpdis->hDC, lpdis->rcItem.left+21+offs, lpdis->rcItem.top, (ETO_CLIPPED|ETO_OPAQUE), &(lpdis->rcItem), buffer, strlen(buffer), NULL);

	   /* Get bitmap size */
	   hBmp = LoadBitmap(hThisInstance, "CLASSBMP");
	   GetObject(hBmp, sizeof(BITMAP), &bm);
	   DeleteObject(hBmp);

	   switch (lpdis->CtlID) {
	     case LB_TYCONS:
		theTycon = (UINT) lpdis->itemID;
		for (i=0, tycons=tyconList; i<theTycon; tycons=tl(tycons))
		    i++;
		tc = hd(tycons);

		switch(tycon(tc).what) {
		    case RESTRICTSYN :
		    case SYNONYM     : hDCMemory = Selected ? hTSDCSel : hTSDC;
				       break;
		    case DATATYPE    : hDCMemory = Selected ? hTCDCSel : hTCDC;
				       break;
		    case NEWTYPE     : hDCMemory = Selected ? hTNDCSel : hTNDC;
				       break;
		}
		break;

	     case LB_CONS:
		if (((UINT)(lpdis->itemID)-2)>=numCfuns)
		    hDCMemory = Selected ? hSDCSel : hSDC;
		else
		    hDCMemory = Selected ? hDDCSel : hDDC;
		break;
	   }

	   BitBlt(lpdis->hDC, (lpdis->rcItem.left)+4+offs, (lpdis->rcItem.top), bm.bmWidth, bm.bmHeight, hDCMemory, 0, 0, SRCCOPY);

	   /* If selected draw rectangle */
	   if ((lpdis->itemState & ODS_SELECTED)&&(lpdis->itemState & ODS_FOCUS)) {
	     DrawFocusRect(lpdis->hDC, &(lpdis->rcItem));
	   }

	   return TRUE;
	 }
	 break;

  case WM_PAINT:

     GetWindowRect(hDlg, &DlgRect);
     GetWindowRect(GetDlgItem(hDlg, ID_BROWSEBMP), &aRect);

     hBitmap = LoadBitmap(hThisInstance, "TYCONSDLGBMP");
     PutBitmap(hBitmap, hDlg, aRect.left-DlgRect.left-GetSystemMetrics(SM_CXDLGFRAME),
	       aRect.top-DlgRect.top-GetSystemMetrics(SM_CYDLGFRAME)-GetSystemMetrics(SM_CYCAPTION),
	       SRCCOPY);
     DeleteObject(hBitmap);
     break;

  case WM_COMMAND:
   switch (wId) {
      case LB_TYCONS:
	switch(NotifyCode) {
	   case LBN_SELCHANGE: {
	     /* A new tycon was selected */
	     theTycon = (UINT) SendDlgItemMessage(hDlg, LB_TYCONS ,LB_GETCURSEL, 0, 0L);
	     SetTycon(hDlg, theTycon, tyconList);
	   }
	   break;

	   case LBN_DBLCLK: {
	     /* Open in text editor script file with tycon definition */
	     CHAR   TyconStr[300];
	     INT    TheTycon;
	     Tycon  tc;

	     /* Get selected tycon */
	     TheTycon = (UINT) SendDlgItemMessage(hDlg, LB_TYCONS ,LB_GETCURSEL, 0, 0L);
	     SendDlgItemMessage(hDlg, LB_TYCONS, LB_GETTEXT, TheTycon, (LPARAM) ((LPSTR) TyconStr));

	     /* Look for Tycon in table */
	     tc = findTycon(findText(TyconStr));

	     if (tycon(tc).line) {
	       module = startNewModule();
	       setLastEdit(scriptName[moduleThisTycon(tc)], tycon(tc).line);
	       dropModulesFrom(module);
	       runEditor();
	     }
	     else {
	       MessageBox(hDlg, "Primitive type:\nNo definition available.", appName, MB_ICONINFORMATION | MB_OK);
	     }
	   }
	   break;
	}
	break;

      case IDC_SEARCHTYCON:  /* Search a name */
	switch(HIBYTE(wId)) {
	  case HIBYTE(EN_CHANGE): {
	    /* Get edit control contents */
	    SendDlgItemMessage(hDlg, IDC_SEARCHTYCON, WM_GETTEXT, MAXBUF, (LPARAM) ((LPSTR) buffer));

	    /* Search in names list box */
	    SendDlgItemMessage(hDlg, LB_TYCONS, LB_SELECTSTRING, 0, (LPARAM) ((LPSTR) buffer));

	    /* Update window contents */
	    DlgSendMessage(hDlg, WM_COMMAND, LB_TYCONS, MAKELONG(0, LBN_SELCHANGE));
	  }
	  break;
	}
	break;

      case ID_EDITTYCON: /* Pushed on Edit tycon button */
	if (SendDlgItemMessage(hDlg, LB_TYCONS, LB_GETCURSEL, 0, 0L) != LB_ERR)
	 DlgSendMessage(hDlg, WM_COMMAND, LB_TYCONS, MAKELONG(0, LBN_DBLCLK));
	break;

      case IDCANCEL:  /* Close dialog */
      case IDOK:
	EndDialog(hDlg, TRUE);
	return TRUE;

      default:
	return TRUE;
   }
 }
 return FALSE;
}

/*-----------------------------------------------------------------------------
 * To draw the classes hierarchy a list with the position of each class is
 * constructed when the window is created (WM_CREATE message). When a WM_PAINT
 * message has to be processed the classes are drawn and joint with lines
 *----------------------------------------------------------------------------*/

/* structure used to draw class hierarchy */
struct HierarchyInfo {
 RECT   Pos;
 Class  Class;
};

static struct HierarchyInfo *Nodes=NULL; /* The list of nodes               */
static INT    LastNode=0;

#define VERTICAL_SEPARATION      35
#define HORIZONTAL_SEPARATION    50
#define INIT_COLUMN              10
#define INIT_ROW		 20

/* Size of class hierarchy window */
static INT  ClassesWidth, ClassesHeight;

/* Expand a node and all its childs recursive */
/* returns the row for placing next node      */
INT Expand(HDC hDC, INT Column, INT Row, Class ThisClass, INT WhatChildAmI)
{
 SIZE  Size;
 INT   i, j, MyPosInList;
 Cell  OneParent;
 List  supers;
 INT   ChildNumber;
 INT   MaxDepth;

 /* Get Size of class name on the screen */
 fprintf(stdstr, "%s\n",textToStr(cclass(ThisClass).text));
 GetTextExtentPoint(hDC,(LPSTR)stdstrbuff, (INT)strlen(stdstrbuff), &Size);

 /* Check if this class is already added */
 for(i=0; i<LastNode; i++)
  if (Nodes[i].Class == ThisClass)
   goto ExpandChilds;

 /* First child at the same vertical position that parent */
 if (WhatChildAmI>1)
  Row++;

 AddParent:

  MyPosInList = LastNode;

  Nodes[MyPosInList].Class      = ThisClass;
  Nodes[MyPosInList].Pos.left   = Column;
  Nodes[MyPosInList].Pos.top    = (Row-1)*VERTICAL_SEPARATION + INIT_ROW;
  Nodes[MyPosInList].Pos.right  = Nodes[MyPosInList].Pos.left + Size.cx;
  Nodes[MyPosInList].Pos.bottom = Nodes[MyPosInList].Pos.top + Size.cy;

  LastNode++;

 ExpandChilds:

 /* Expand parent childs */
 ChildNumber = 0;
 for(i=CLASSMIN; i<classMax(); i++) {

   /* Check for cycles in graph */
   for(j=0; j<LastNode; j++)
    if(Nodes[j].Class == i)
     goto TakeOtherParent;

   for(supers=cclass(i).supers; nonNull(supers); supers=tl(supers)) {
     /* Get a parent for the class */
     OneParent = hd(supers);
     /* If parent is the same that this class, expand this class (it is a child) */
     if (OneParent == ThisClass) {
       Row = Expand(hDC, Nodes[MyPosInList].Pos.right+HORIZONTAL_SEPARATION, Row, i, ++ChildNumber);
     }
   }
 TakeOtherParent:
 }

 /* This code corrects the initial vertical position of the expanded class */
 /* using childs vertical positions                                        */
 MaxDepth=0;

 for(i=0; i<LastNode; i++)
  if (Nodes[i].Pos.bottom > MaxDepth)
    MaxDepth = Nodes[i].Pos.bottom;

 Nodes[MyPosInList].Pos.top = Nodes[MyPosInList].Pos.top+((MaxDepth - Nodes[MyPosInList].Pos.bottom)/2);
 Nodes[MyPosInList].Pos.bottom = Nodes[MyPosInList].Pos.top + Size.cy;

 return Row;
}


#define MAX_WIDTH  		600
#define MAX_HEIGHT 		400


/* Hierarchy class window proc */
LRESULT FAR APIENTRY _export ClassesWndProc     (HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{

 static HCURSOR 	hMoveClassCursor, hNormalCursor;
 static INT 		SelectedClass = -1;
 static BOOL 		Moved=FALSE;
 static INT 		ClassesTopX=10, ClassesTopY=10;
 static INT 		XOffset, YOffset;
 static INT 		RealWidth, RealHeight;

 switch (msg) {

  case WM_CREATE: {
   PAINTSTRUCT ps;
   HDC         hDC;
   INT         i, j, k, numClasses, Row;
   HFONT       hFont=NULL, hSaveFont;
   List        supers;
   Cell	       OneParent;
   BOOL        GoBackToCycle;


   /* Get the total number of classes */
   numClasses = classMax() - CLASSMIN;

   /* Get memory for nodes list */
#if BCWIN16
   if (Nodes) farfree(Nodes);
   Nodes = farcalloc((ULONG)(sizeof(struct HierarchyInfo)),
                     (ULONG) numClasses);
#else
   if (Nodes) free(Nodes);
   Nodes = calloc((ULONG)(sizeof(struct HierarchyInfo)),
                  (ULONG) numClasses);

#endif

   LastNode = 0;

   if (!Nodes) {
     MessageBox(hWnd, "Out of memory: create nodes list", NULL, MB_ICONEXCLAMATION | MB_OK);
     break;
   }

   hDC = BeginPaint(hWnd, &ps);

   /* Get a text font */
   hFont = CreateFont( 15,
			7,
			0,
			0,
			FW_BOLD,
			FALSE,
			FALSE,
			FALSE,
			ANSI_CHARSET,
			OUT_TT_PRECIS,
			CLIP_TT_ALWAYS,
			DEFAULT_QUALITY,
			FIXED_PITCH | FF_MODERN,
			"Arial New");

   if (hFont)
     hSaveFont = SelectObject(hDC, hFont);

   /* Find a class with no superclass and expand it */
   Row = 0;
   for(i=CLASSMIN; i<classMax(); i++)
     if (cclass(i).numSupers == 0) {
       Row = Expand(hDC, INIT_COLUMN, ++Row, i, 0);
       break;
     }

   TheCycle:

     if (LastNode==numClasses)
       goto End;

     /* Look for superclassses wich childs are already drawn */
     do {
       GoBackToCycle = FALSE;
       for(i=0; i<LastNode; i++) {
	 if (cclass(Nodes[i].Class).numSupers) {
	   for(supers=cclass(Nodes[i].Class).supers; nonNull(supers); supers=tl(supers)) {
	     OneParent = hd(supers);

	     if (cclass(OneParent).numSupers == 0) {
	       for(k=0; k<LastNode; k++)
	       if (Nodes[k].Class == OneParent)
	       goto AlreadyExpanded;

	       Row = Expand(hDC, INIT_COLUMN, ++Row, OneParent, 0);
	       GoBackToCycle = TRUE;
	     }

	     AlreadyExpanded:
	   }
	 }
       }
     } while (GoBackToCycle);


     /* Find other supeclass, expand it and expand all the */
     /* classes related to it                              */
     for(i=CLASSMIN; i<classMax(); i++) {
       if (cclass(i).numSupers == 0) {
	 for(j=0; j<LastNode; j++)
	 if (Nodes[j].Class == i)
	   goto AlreadyExpanded2;

	 Row = Expand(hDC, INIT_COLUMN, ++Row, i, 0);
	 goto TheCycle;
       }
       AlreadyExpanded2:
     }

     /* Look for any other class and expand it */
     /* Expand also classes related to it      */
     for(i=CLASSMIN; i<classMax(); i++) {
       for(j=0; j<LastNode; j++)
	 if (Nodes[j].Class == i)
	   goto AlreadyExpanded3;

       Row = Expand(hDC, INIT_COLUMN, ++Row, i, 0);
       goto TheCycle;

       AlreadyExpanded3:
     }

     End:

     /* Restore font */
     if (hFont) {
       SelectObject(hDC, hSaveFont);
       DeleteObject(hFont);
     }

     EndPaint(hWnd, &ps);

     /* Load cursors */
     hNormalCursor = LoadCursor(NULL, IDC_ARROW);
     hMoveClassCursor = LoadCursor(hThisInstance, "MOVECLASSCURSOR");

     /* Show upper-left part of window */
     XOffset = YOffset = 0;
     SetScrollPos(hWnd, SB_HORZ, 0, TRUE);
     SetScrollPos(hWnd, SB_VERT, 0, TRUE);

     /* Get window size */
     ChangeSize:
     for (i=0, ClassesWidth=ClassesHeight=0; i<LastNode; i++) {
       if (Nodes[i].Pos.right > ClassesWidth)
	 ClassesWidth = Nodes[i].Pos.right;
	 if (Nodes[i].Pos.bottom > ClassesHeight)
	   ClassesHeight = Nodes[i].Pos.bottom;
     }

     ClassesWidth  += 2*GetSystemMetrics(SM_CXFRAME);
     ClassesHeight += 2*GetSystemMetrics(SM_CYFRAME)+
		      GetSystemMetrics(SM_CYCAPTION);

     MoveWindow (hWndClasses, ClassesTopX, ClassesTopY,
		 ClassesWidth+GetSystemMetrics(SM_CXVSCROLL)+10,
		 ClassesHeight+GetSystemMetrics(SM_CYHSCROLL)+10, TRUE);

     break;

  }

  case WM_DESTROY:
#if BCWIN16
   if (Nodes) farfree(Nodes);
#else
   if (Nodes) free(Nodes);
#endif
   Nodes = NULL;
   LastNode = 0;
   DestroyCursor(hMoveClassCursor);
   hWndClasses=NULL;

   return (LRESULT) FALSE;

  case WM_GETMINMAXINFO: {

   MINMAXINFO *lpmmi;

   lpmmi = (MINMAXINFO FAR*) lParam;

   lpmmi->ptMinTrackSize.x = 50;
   lpmmi->ptMinTrackSize.y = 50;

   lpmmi->ptMaxTrackSize.x = MAX_WIDTH;
   lpmmi->ptMaxTrackSize.y = MAX_HEIGHT;

   lpmmi->ptMaxSize.x = MAX_WIDTH;
   lpmmi->ptMaxSize.y = MAX_HEIGHT;

   break;
  }

  case WM_SIZE: {

   static BOOL AvoidRecursion=FALSE;

   if (AvoidRecursion)
     break;

   AvoidRecursion = TRUE;

   RealWidth = LOWORD(lParam);
   RealHeight = HIWORD(lParam);

   if (RealWidth < ClassesWidth || XOffset) {
     SetScrollRange(hWnd, SB_HORZ, 0, ClassesWidth, TRUE);
   }
   else {
     /* Hide scroll bar */
     SetScrollRange(hWnd, SB_HORZ, 0, 0, TRUE);
   }

   if (RealHeight < ClassesHeight || YOffset) {
     SetScrollRange(hWnd, SB_VERT, 0, ClassesHeight, TRUE);
   }
   else {
     /* Hide scroll bar */
     SetScrollRange(hWnd, SB_VERT, 0, 0, TRUE);
   }

   AvoidRecursion = FALSE;

   break;
  }

  case WM_MOVE:

   /* We must subtract as it gets client position */
   ClassesTopX = (INT) LOWORD(lParam) - GetSystemMetrics(SM_CXFRAME);
   ClassesTopY = (INT) HIWORD(lParam) - GetSystemMetrics(SM_CYCAPTION)
				      - GetSystemMetrics(SM_CYFRAME)
				      + GetSystemMetrics(SM_CYBORDER);
   break;

  case WM_HSCROLL: {

   INT i;

   switch(LOWORD(wParam)) {

     case SB_PAGEUP:
     case SB_LINEUP:
       if (XOffset < 0) {
	 XOffset+=5;
	 SetScrollPos(hWnd, SB_HORZ, -XOffset, TRUE);
	 for (i=0; i<LastNode; i++) {
	   Nodes[i].Pos.left+=5;
	   Nodes[i].Pos.right+=5;
	 }
	 ScrollWindow(hWnd, 5, 0, NULL, NULL);
	 UpdateWindow(hWnd);
       }
       break;

     case SB_PAGEDOWN:
     case SB_LINEDOWN:
       if (-XOffset < ClassesWidth) {
	 XOffset-=5;
	 SetScrollPos(hWnd, SB_HORZ, -XOffset, TRUE);
	 for (i=0; i<LastNode; i++) {
	   Nodes[i].Pos.left-=5;
	   Nodes[i].Pos.right-=5;
	 }
	 ScrollWindow(hWnd, -5, 0, NULL, NULL);
	 UpdateWindow(hWnd);
       }
       break;

     case SB_THUMBPOSITION: {

#if BCWIN32
       INT NewPosition = HIWORD(wParam);
#else
       INT NewPosition = LOWORD(lParam);
#endif

       SetScrollPos(hWnd, SB_HORZ, NewPosition, TRUE);
       for (i=0; i<LastNode; i++) {
	 Nodes[i].Pos.left-=(XOffset+NewPosition);
	 Nodes[i].Pos.right-=(XOffset+NewPosition);
       }
       XOffset = -NewPosition;
       InvalidateRect(hWnd,NULL,TRUE);
       UpdateWindow(hWnd);
     }
     break;

   }

  }
  break;

  case WM_VSCROLL: {

   INT i;

   switch(LOWORD(wParam)) {

     case SB_PAGEUP:
     case SB_LINEUP:
       if (YOffset < 0) {
	 YOffset+=5;
	 SetScrollPos(hWnd, SB_VERT, -YOffset, TRUE);
	 for (i=0; i<LastNode; i++) {
	   Nodes[i].Pos.top+=5;
	   Nodes[i].Pos.bottom+=5;
	 }
	 ScrollWindow(hWnd, 0, 5, NULL, NULL);
	 UpdateWindow(hWnd);
       }
       break;

     case SB_PAGEDOWN:
     case SB_LINEDOWN:
       if (-YOffset < ClassesHeight) {
	 YOffset-=5;
	 SetScrollPos(hWnd, SB_VERT, -YOffset, TRUE);
	 for (i=0; i<LastNode; i++) {
	   Nodes[i].Pos.top-=5;
	   Nodes[i].Pos.bottom-=5;
	 }
	 ScrollWindow(hWnd, 0, -5, NULL, NULL);
	 UpdateWindow(hWnd);
       }
       break;

     case SB_THUMBPOSITION: {

#if BCWIN32
       INT NewPosition = HIWORD(wParam);
#else
       INT NewPosition = LOWORD(lParam);
#endif

       SetScrollPos(hWnd, SB_VERT, NewPosition, TRUE);
       for (i=0; i<LastNode; i++) {
	 Nodes[i].Pos.top-=(YOffset+NewPosition);
	 Nodes[i].Pos.bottom-=(YOffset+NewPosition);
       }
       YOffset = -NewPosition;
       InvalidateRect(hWnd,NULL,TRUE);
       UpdateWindow(hWnd);
     }
     break;
   }
  }
  break;

  case WM_PAINT: {

    PAINTSTRUCT ps;
    HDC         hDC;
    INT         i, OriginClass, TargetClass;
    HFONT       hFont=NULL, hSaveFont;
    List        supers;
    COLORREF    SaveColor;
    Cell        OneParent;

    hDC = BeginPaint(hWnd, &ps);

    /* If dragging a class don't paint joint lines */
    if (SelectedClass < 0) {

      /* Paint joint lines */
      for(i=CLASSMIN; i<classMax(); i++)
	 for(supers=cclass(i).supers; nonNull(supers); supers=tl(supers)) {
	   OneParent = hd(supers);

	    /* If it is a child of itself paint a line to itself */
	    if (OneParent == i) {
	       /* Find class in List */
	       for(OriginClass=0; Nodes[OriginClass].Class != i; OriginClass++);
	       Arc(hDC, Nodes[OriginClass].Pos.right-5,  Nodes[OriginClass].Pos.bottom-5,
			Nodes[OriginClass].Pos.right+15, Nodes[OriginClass].Pos.bottom+20,
			Nodes[OriginClass].Pos.right-5,  Nodes[OriginClass].Pos.bottom-5,
			Nodes[OriginClass].Pos.right-4,  Nodes[OriginClass].Pos.bottom-4);

	    }
	    else {
	      /* Joint the two classes with a line */

	      /* Find the two classses to joint */
	      for(OriginClass=0; Nodes[OriginClass].Class != i; OriginClass++);
	      for(TargetClass=0; Nodes[TargetClass].Class != OneParent; TargetClass++);

              if (Nodes[OriginClass].Pos.left > Nodes[TargetClass].Pos.right) {
                MoveToEx(hDC, Nodes[OriginClass].Pos.left-4,
                              Nodes[OriginClass].Pos.top+(Nodes[OriginClass].Pos.bottom-Nodes[OriginClass].Pos.top)/2,
                              NULL);
                LineTo(hDC, Nodes[TargetClass].Pos.right+4,
			    Nodes[TargetClass].Pos.top+(Nodes[TargetClass].Pos.bottom-Nodes[TargetClass].Pos.top)/2);
	      }
              else {
                MoveToEx(hDC, Nodes[TargetClass].Pos.left-4,
			      Nodes[TargetClass].Pos.top+(Nodes[TargetClass].Pos.bottom-Nodes[TargetClass].Pos.top)/2,
                              NULL);
                LineTo(hDC, Nodes[OriginClass].Pos.right+4,
			    Nodes[OriginClass].Pos.top+(Nodes[OriginClass].Pos.bottom-Nodes[OriginClass].Pos.top)/2);
	      }
	    }
	   }

    }

    /* Get font */
    hFont = CreateFont( 15,
			7,
			0,
			0,
			FW_BOLD,
			FALSE,
			FALSE,
			FALSE,
			ANSI_CHARSET,
			OUT_TT_PRECIS,
			CLIP_TT_ALWAYS,
			DEFAULT_QUALITY,
			FIXED_PITCH | FF_MODERN,
			"Arial New");

    if (hFont)
      hSaveFont = SelectObject(hDC, hFont);

    /* Blue Color for text */
    SaveColor = SetTextColor(hDC, RGB(0,0,190));

    /* Draw class text and frames */
    for(i=0; i<LastNode; i++) {

      /* frame */
      Rectangle(hDC, Nodes[i].Pos.left-4, Nodes[i].Pos.top-2,
		     Nodes[i].Pos.right+4,Nodes[i].Pos.bottom+2);

      /* frame shadow */
      MoveToEx(hDC, Nodes[i].Pos.right+4, Nodes[i].Pos.top, NULL);
      LineTo(hDC, Nodes[i].Pos.right+4, Nodes[i].Pos.bottom+2);
      LineTo(hDC, Nodes[i].Pos.left-2, Nodes[i].Pos.bottom+2);

      /* class text */
      fprintf(stdstr, "%s\n",textToStr(cclass(Nodes[i].Class).text));
      TextOut(hDC, Nodes[i].Pos.left, Nodes[i].Pos.top, (LPSTR)stdstrbuff,
	      (INT)strlen(stdstrbuff));

    }

    /* Restore color */
    SetTextColor(hDC, SaveColor);

    /* Restore font */
    if (hFont) {
      SelectObject(hDC, hSaveFont);
      DeleteObject(hFont);
    }

    EndPaint(hWnd, &ps);

    break;
  }

  case WM_LBUTTONDOWN: {

    /* Select a class to drag it */
    INT i, x, y;

    x = LOWORD(lParam);
    y = HIWORD(lParam);


    for(i=0; i<LastNode; i++)
    if (Nodes[i].Pos.left-4 < x && Nodes[i].Pos.right+4  > x &&
	Nodes[i].Pos.top-2  < y && Nodes[i].Pos.bottom+2 > y) {

      SetCursor(hMoveClassCursor);

      SelectedClass = i;

      /* Clear window */
      InvalidateRect(hWnd, NULL, TRUE);

      SetCapture(hWnd);
    }
    break;
  }

  case WM_LBUTTONUP: {

    INT x, y, width, height;

    if (SelectedClass < 0)
      break;

    x = LOWORD(lParam);
    y = HIWORD(lParam);

    width = Nodes[SelectedClass].Pos.right - Nodes[SelectedClass].Pos.left;
    height = Nodes[SelectedClass].Pos.bottom - Nodes[SelectedClass].Pos.top;

    ReleaseCapture();

    if (Moved) {
      Nodes[SelectedClass].Pos.left = x;
      Nodes[SelectedClass].Pos.top =  y;
      Nodes[SelectedClass].Pos.right = Nodes[SelectedClass].Pos.left+width;
      Nodes[SelectedClass].Pos.bottom = Nodes[SelectedClass].Pos.top+height;
    }

    SelectedClass = -1;
    Moved = FALSE;
    SetCursor(hNormalCursor);

    InvalidateRect(hWnd, NULL, TRUE);
    SendMessage(hWnd, WM_PAINT, 0, 0L);

    goto ChangeSize;

  }

  case WM_MOUSEMOVE:
    if (SelectedClass < 0)
      SetCursor(hNormalCursor);
    else {

      INT  x, y, width, height, saveX, saveY;
      RECT ClearRect;

      SetCursor(hMoveClassCursor);

      /* Don't allow move it out of window */
      lParam = MAKELONG(min(RealWidth-10, (INT)LOWORD(lParam)),
			min(RealHeight-10,(INT)HIWORD(lParam)));
      lParam = MAKELONG(max(5, (INT)LOWORD(lParam)),
			max(5, (INT)HIWORD(lParam)));

      x = LOWORD(lParam);
      y = HIWORD(lParam);

      width = Nodes[SelectedClass].Pos.right - Nodes[SelectedClass].Pos.left;
      height = Nodes[SelectedClass].Pos.bottom - Nodes[SelectedClass].Pos.top;

      saveX = Nodes[SelectedClass].Pos.left;
      saveY = Nodes[SelectedClass].Pos.top;

      Nodes[SelectedClass].Pos.left = x;
      Nodes[SelectedClass].Pos.top =  y;
      Nodes[SelectedClass].Pos.right = Nodes[SelectedClass].Pos.left+width;
      Nodes[SelectedClass].Pos.bottom = Nodes[SelectedClass].Pos.top+height;

      if (x > saveX) {
	/* class has moved left */
	ClearRect.left = saveX-5;
	ClearRect.right = Nodes[SelectedClass].Pos.right+5;
      }
      else {
	ClearRect.left = x-5;
	ClearRect.right = saveX+width+5;
      }

      if (y > saveY) {
	/* class moved down */
	ClearRect.top = saveY-3;
	ClearRect.bottom = Nodes[SelectedClass].Pos.bottom+3;
      }
      else {
	ClearRect.top = Nodes[SelectedClass].Pos.top-3;
	ClearRect.bottom = saveY+height+3;
      }

      InvalidateRect(hWnd, &ClearRect, TRUE);
      SendMessage(hWnd, WM_PAINT, 0, 0L);

    }
    break;

  default:
    return DefWindowProc(hWnd, msg, wParam, lParam);
 }
 return (LONG) TRUE;
}


/* Create class hierarchy and show it on a window */
static VOID local Hierarchy(VOID)
{
 HWND hActiveWindow;
 RECT rActive, rWindow;

 if (hWndClasses) {
  /* If window exits keep its position */
  GetWindowRect(hWndClasses, &rWindow);
  DestroyWindow(hWndClasses);
  hActiveWindow = GetActiveWindow();
 }
 else {
  hActiveWindow = GetActiveWindow();
  GetWindowRect(hActiveWindow, &rActive);
  rWindow.top = rActive.top + 50;
  rWindow.left = rActive.left + 50;
 }

 hActiveWindow = GetActiveWindow();

 hWndClasses =
  CreateWindow("HugsClassesWindow",
	       "Class Hierarchy",
	       WS_CAPTION | WS_BORDER | WS_SYSMENU | WS_THICKFRAME | WS_VSCROLL | WS_HSCROLL,
	       0, 0, 0, 0,
	       (HWND) hActiveWindow,
	       (HMENU) NULL,
	       hThisInstance,
	       (LPSTR) NULL);

 if (!hWndClasses) {
   MessageBox (hWndMain, "Error creating window", appName, MB_ICONEXCLAMATION | MB_OK);
   return;
 }

 MoveWindow(hWndClasses, rWindow.left, rWindow.top,
	    min(ClassesWidth+GetSystemMetrics(SM_CXVSCROLL)+10,  MAX_WIDTH),
	    min(ClassesHeight+GetSystemMetrics(SM_CYHSCROLL)+10, MAX_HEIGHT), TRUE);

 ShowWindow  (hWndClasses, SW_SHOWNORMAL);
 UpdateWindow(hWndClasses);

 SetFocus(hWndClasses);

 return;
}


